from __future__ import annotations

import logging
import os

import click

import vpype as vp

from .cli import cli
from .decorators import global_processor, pass_state
from .state import State
from .types import FileType, IntegerType, TextType

__all__ = ("write",)

WRITE_HELP = f"""Save geometries to a file.

The `write` command support two format: SVG and HPGL. The format is determined based on the
file extension used for `OUTPUT` or the `--file-format` option. The latter is useful when
`OUTPUT` is a single dash (`-`), in which case the output is printed to stdout instead of a
file.

When writing to SVG, the current page size is used if available. The current page size is
implicitly set by the `read` command and can also be manually changed using the `pagesize`
command. The page size can be overridden using the `--page-size SIZE` option. `SIZE` may
be one of:

    {', '.join(vp.PAGE_SIZES.keys())}

Alternatively, a custom size can be specified in the form of `WIDTHxHEIGHT`. `WIDTH` and
`HEIGHT` may include units. If only one has an unit, the other is assumed to have the
same unit. If none have units, both are assumed to be pixels by default. Here are some
examples:

\b
    --page-size 11x14in     # 11in by 14in
    --page-size 1024x768    # 1024px by 768px
    --page-size 13.5inx4cm  # 13.5in by 4cm

When a page size is provided, it will be rotated if the `--landscape` option is used.

If the current page set has not been set (e.g. because the `read` command has not been used)
and the `--page-size` is not provided, the SVG will have its bounds tightly fit to the
geometries.

By default, the geometries are not scaled or translated even if they lie outside of the page
boundaries. The `--center` option translates the geometries to the center of the page.

Layers are labelled with their numbers by default. If an alternative naming is required, a
template pattern can be provided using the `--layer-label` option. The provided pattern must
contain a C-style format specifier such as `%d` which will be replaced by the layer number.

By default, paths are colored according to the corresponding layer property (as set by the 
`color` or `read` commands). If the color property is not set, a default, per-layer color
scheme is used. Alternative behaviours are available with the `--color-mode` option. Setting it
to "none" disables coloring and black paths are generated. Setting it to "layer" applies the
default color scheme to each layer. Setting it to "path" gives a different color to each path
(with a rotation), which makes it possible to visualize line optimization.

Stroke widths are set according to the corresponding property (as et by the `penwdith`, `pen`,
or `read` commands). If the property is missing, a 1px default is used.

Pen-up trajectories can be generated with the `--pen-up` flag. As most plotting tools will
include these paths in the output, this option should be used for previsualisation only. The
Axidraw tools will however ignore them.

If the `--restore-attribs` option is used, the SVG attributes extracted by the `read` command
are restored in the output SVG. (Note that this is an experimental feature which is unable to
fully recreate an input SVG in all but the simplest cases.)

When writing to HPGL, a device name must be provided with the `--device` option. The
corresponding device must be configured in the built-in or a user-provided configuration file
(see the documentation for more details). The following devices are currently available:

    {', '.join(vp.config_manager.get_plotter_list())}

In HPGL mode, this command will try to infer the paper size to use based on the current page
size (the current page size is set by the `read` command based on the input file and can be
manually set or changed with the `pagesize` or `layout` command). An error will be displayed if
no corresponding paper size if found. Use the `--page-size` option with a format defined in the
device's configuration to manually specify with paper size to use.

The plotter may need to be specifically configured for the desired paper size (e.g. for A4 or
A3, the HP 7475a's corresponding DIP switch must be set to metric mode). A note will be
displayed as a reminder and can be hidden using the `--quiet` option.

The `--landscape` and `--center` options are accepted and honored in HPGL.

By default, relative coordinates are used whenever possible in the HPGL output in order to
reduce file size. If `--absolute` is used, absolute coordinate are used instead.

Optionally, the HPGL-only `--velocity` can be provided, in which case a `VS` command will be
emitted with the provided value.

Examples:

    Write to a SVG using the current page size as set by the `read` command:

        vpype read input.svg [...] write output.svg

    Write to a portrait A4 page:

        vpype [...] write --page-size a4 output.svg

    Write to a 13x9 inch page and center the geometries:

        vpype [...] write --page-size 13x9in --landscape --center output.svg

    Use custom layer labels:

        vpype [...] write --page-size a4 --layer-label "Pen %d" output.svg

    Write a previsualization SVG:

        vpype [...] write --pen-up --color-mode path output.svg

    Output SVG to stdout:

        vpype [...] write --format SVG -

    Write a A4 page with portrait orientation HPGL file:

        vpype [...] write --device hp7475a --page-size a4 --center
"""


@cli.command(group="Output", help=WRITE_HELP)
@click.argument("output", type=FileType("w"))
@click.option(
    "-f",
    "--format",
    "file_format",
    type=click.Choice(["svg", "hpgl"], case_sensitive=False),
    help="Output format (inferred from file extension by default).",
)
@click.option(
    "-p",
    "--page-size",
    type=str,
    help=(
        "Set the bounds of the SVG to a specific page size. If omitted, the SVG size is set "
        "to the current page size (see `read` and `pagesize` commands. May not be omitted for "
        "HPGL."
    ),
)
@click.option(
    "-l", "--landscape", is_flag=True, help="Use landscape orientation instead of portrait."
)
@click.option(
    "-c", "--center", is_flag=True, help="Center the geometries within the SVG bounds."
)
@click.option(
    "-ll",
    "--layer-label",
    type=TextType(),
    help="[SVG only] Pattern used to for naming layers.",
)
@click.option("-pu", "--pen-up", is_flag=True, help="[SVG only] Generate pen-up trajectories.")
@click.option(
    "-m",
    "--color-mode",
    type=click.Choice(["default", "none", "layer", "path"]),
    default="default",
    help="[SVG only] Color mode for paths (default: default).",
)
@click.option(
    "-r",
    "--restore-attribs",
    is_flag=True,
    default=False,
    help="[SVG only] attempt to restore SVG attributes from properties.",
)
@click.option(
    "--dont-set-date",
    is_flag=True,
    help="[SVG only] do not add date metadata (useful for auto-generated SVG under VCS).",
)
@click.option(
    "-d", "--device", type=TextType(), help="[HPGL only] Type of the plotter device."
)
@click.option(
    "-a", "--absolute", is_flag=True, help="[HPGL only] Use absolute coordinates exclusively."
)
@click.option(
    "-vs",
    "--velocity",
    type=IntegerType(),
    help="[HPGL only] Emit a VS command with the provided value.",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="[HPGL only] Do not display the plotter configuration or paper loading information.",
)
@click.pass_obj  # to obtain the command string
@global_processor
@pass_state
def write(
    state: State,
    document: vp.Document,
    cmd_string: str | None,
    output,
    file_format: str,
    page_size: str,
    landscape: bool,
    center: bool,
    layer_label: str | None,
    restore_attribs: bool,
    dont_set_date: bool,
    pen_up: bool,
    color_mode: str,
    device: str | None,
    absolute: bool,
    velocity: int | None,
    quiet: bool,
):
    """Write command."""

    if document.is_empty():
        logging.warning("no geometry to save, no file created")
        return document

    if file_format is None:
        # infer format
        _, ext = os.path.splitext(output.name)
        file_format = ext.lstrip(".").lower()

    if file_format == "svg":
        page_size_px = None
        if page_size:
            page_size_px = vp.convert_page_size(state.substitute(page_size))
            if landscape:
                page_size_px = page_size_px[::-1]

        vp.write_svg(
            output=output,
            document=document,
            page_size=page_size_px,
            center=center,
            source_string=cmd_string if cmd_string is not None else "",
            layer_label_format=layer_label,
            show_pen_up=pen_up,
            color_mode=color_mode,
            use_svg_metadata=restore_attribs,
            set_date=not dont_set_date,
        )
    elif file_format == "hpgl":
        if not page_size:
            config = vp.config_manager.get_plotter_config(device)
            if config is not None:
                paper_config = config.paper_config_from_size(document.page_size)
            else:
                paper_config = None

            if paper_config and document.page_size is not None:
                page_size = paper_config.name
                landscape = document.page_size[0] > document.page_size[1]
            else:
                logging.error(
                    "write: the plotter page size could not be inferred from the current page "
                    "size (use the `--page-size SIZE` option)"
                )
                return document

        vp.write_hpgl(
            output=output,
            document=document,
            landscape=landscape,
            center=center,
            device=device,
            page_size=page_size,
            absolute=absolute,
            velocity=velocity,
            quiet=quiet,
        )
    else:
        logging.warning(
            f"write: format could not be inferred or format unknown '{file_format}', "
            "no file created (use the `--format` option)"
        )

    return document



from __future__ import annotations

import dataclasses
import pathlib

import svgelements


@dataclasses.dataclass(frozen=True)
class Color:
    """Simple, immutable, hashable color class with flexible construction.

    Examples:

        >>> Color()
        Color(red=0, green=0, blue=0, alpha=255)
        >>> Color(255, 0, 255)
        Color(red=255, green=0, blue=255, alpha=255)
        >>> Color(255, 0, 255, 128)
        Color(red=255, green=0, blue=255, alpha=128)
        >>> red = Color("red")
        >>> red
        Color(red=255, green=0, blue=0, alpha=255)
        >>> Color(red)
        Color(red=255, green=0, blue=0, alpha=255)
        >>> Color(svgelements.Color("#0f0"))
        Color(red=0, green=255, blue=0, alpha=255)
    """

    red: int
    green: int
    blue: int
    alpha: int

    def __init__(
        self,
        red: int | str | Color | svgelements.Color | None = None,
        green: int | None = None,
        blue: int | None = None,
        alpha: int | None = None,
    ):
        svgc = None
        if isinstance(red, (svgelements.Color, Color)):
            svgc = red
        elif isinstance(red, str):
            svgc = svgelements.Color(red)
        if svgc is not None:
            red, green, blue, alpha = svgc.red, svgc.green, svgc.blue, svgc.alpha

        object.__setattr__(self, "red", int(red or 0))  # type: ignore
        object.__setattr__(self, "green", int(green or 0))
        object.__setattr__(self, "blue", int(blue or 0))
        object.__setattr__(self, "alpha", int(alpha if alpha is not None else 255))

    def as_floats(self) -> tuple[float, float, float, float]:
        """Returns a float representation of the instance."""
        return self.red / 255, self.green / 255, self.blue / 255, self.alpha / 255

    def as_hex(self) -> str:
        """Return a standard, hexadecimal representation of the instance."""
        return svgelements.Color(self.red, self.green, self.blue, self.alpha).hex

    def as_rgb_hex(self) -> str:
        """Return a standard, hexadecimal representation of the instance, ignoring alpha."""
        return svgelements.Color(self.red, self.green, self.blue).hexrgb

    def __str__(self) -> str:
        return self.as_hex()


# layer metadata field names
METADATA_FIELD_NAME = "vp_name"
METADATA_FIELD_COLOR = "vp_color"
METADATA_FIELD_PEN_WIDTH = "vp_pen_width"
METADATA_FIELD_SOURCE = "vp_source"
METADATA_FIELD_SOURCE_LIST = "vp_sources"

# global metadata field names
METADATA_FIELD_SVG_NAMESPACES = "vp_svg_ns"
METADATA_FIELD_PAGE_SIZE = "vp_page_size"

METADATA_SYSTEM_FIELD_TYPES = {
    METADATA_FIELD_NAME: str,
    METADATA_FIELD_COLOR: Color,
    METADATA_FIELD_PEN_WIDTH: float,
    METADATA_FIELD_PAGE_SIZE: tuple,
    METADATA_FIELD_SOURCE: pathlib.Path,
    METADATA_FIELD_SOURCE_LIST: set,
}

# noinspection HttpUrlsUsage
METADATA_SVG_NAMESPACES = {
    "http://www.inkscape.org/namespaces/inkscape": "inkscape",
    "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd": "sodipodi",
    "http://purl.org/dc/elements/1.1/": "dc",
    "http://creativecommons.org/ns": "cc",
    "http://www.w3.org/1999/02/22-rdf-syntax-ns": "rdf",
}

METADATA_SVG_ATTRIBUTES_WHITELIST = {
    # list based on https://css-tricks.com/svg-properties-and-css/
    # font properties
    "font",
    "font-family",
    "font-size",
    "font-size-adjust",
    "font-stretch",
    "font-style",
    "font-variant",
    "font-weight",
    # text properties
    "direction",
    "letter-spacing",
    "text-decoration",
    "unicode-bidi",
    "word-spacing",
    "writing-mode",
    "alignment-baseline",
    "baseline-shift",
    "dominant-baseline",
    "glyph-orientation-horizontal",
    "glyph-orientation-vertical",
    "kerning",
    "text-anchor",
    # masking properties
    "overflow",
    "mask",
    "opacity",
    # filter effect
    "enable-background",
    "filter",
    # interactivity properties
    "cursor",
    "pointer-events",
    # visibility properties
    "display",
    "visibility",
    # painting properties
    "color-interpolation",
    "color-interpolation-filters",
    "color-rendering",
    "fill",
    "fill-rule",
    "fill-opacity",
    "image-rendering",
    "marker",
    "marker-start",
    "marker-mid",
    "marker-end",
    "shape-rendering",
    "stroke",
    "stroke-dasharray",
    "stroke-dashoffset",
    "stroke-linecap",
    "stroke-linejoin",
    "stroke-miterlimit",
    "stroke-opacity",
    "stroke-width",
    "text-rendering",
}


METADATA_DEFAULT_COLOR_SCHEME = [
    Color("#00f"),
    Color("#080"),
    Color("#f00"),
    Color("#0cc"),
    Color("#0f0"),
    Color("#c0c"),
    Color("#cc0"),
    Color("black"),
]



from __future__ import annotations

import logging
from typing import Any, Callable

import click

import vpype as vp

from .cli import cli
from .decorators import global_processor, layer_processor
from .types import FloatRangeType, LayerType, LengthType, TextType, multiple_to_layer_ids

__all__ = (
    "propset",
    "proplist",
    "propget",
    "propdel",
    "propclear",
    "penwidth",
    "color",
    "alpha",
    "name",
    "pens",
)


_STR_TO_TYPE: dict[str, Callable] = {
    "str": str,
    "int": lambda x: int(vp.convert_length(x)),
    "float": vp.convert_length,
    "color": vp.Color,
}


def _check_scope(
    global_flag: bool, layer: int | list[int] | None
) -> tuple[bool, int | list[int] | None]:
    if global_flag and layer is not None:
        logging.warning(
            "incompatible `--global` and `--layer` options were provided, assuming `--global`"
        )
        layer = None
    elif not global_flag and layer is None:
        logging.warning(
            "neither `--global` nor `--layer` options were provide, assuming `--layer all`"
        )
        layer = LayerType.ALL

    return global_flag, layer


@cli.command(group="Metadata")
@click.argument("prop", type=TextType())
@click.argument("value", type=TextType())
@click.option("--global", "-g", "global_flag", is_flag=True, help="Global mode.")
@click.option("-l", "--layer", type=LayerType(accept_multiple=True), help="Target layer(s).")
@click.option(
    "--type",
    "-t",
    "prop_type",
    type=click.Choice(list(_STR_TO_TYPE.keys())),
    default="str",
    help="Property type.",
)
@global_processor
def propset(
    document: vp.Document,
    global_flag: bool,
    layer: int | list[int] | None,
    prop: str,
    value: str,
    prop_type: str,
):
    """Set the value of a global or layer property.

    Either the `--global` or `--layer` option must be used to specify whether a global or
    layer property should be set. When using `--layer`, either a single layer ID, a
    coma-separated list of layer ID, or `all` may be used.

    By default, the value is stored as a string. Alternative types may be specified with the
    `--type` option. The following types are available:

        \b
        str: text string
        int: integer number
        float: floating-point number
        color: color

    When using the `int` and `float` types, units may be used and the value will be converted
    to pixels.

    When using the `color` type, any SVG-compatible string may be used for VALUE, including
    16-bit RGB (#ff0000), 16-bit RGBA (#ff0000ff), 8-bit variants (#f00 or #f00f), or color
    names (red).

    Examples:

        Set a global property of type `int`:

            vpype [...] propset --global --type int my_prop 10 [...]

        Set the layer property of type `float` (this is equivalent to using the `penwidth`
        command:

            vpype [...] propset --layer 1 --type float vp_pen_width 0.5mm [...]

        Set a layer property of type `color`:

            vpype [...] propset --layer 1 --type color my_prop red [...]
    """

    global_flag, layer = _check_scope(global_flag, layer)

    if global_flag:
        document.set_property(prop, _STR_TO_TYPE[prop_type](value))
    else:
        for lid in multiple_to_layer_ids(layer, document):
            document.layers[lid].set_property(prop, _STR_TO_TYPE[prop_type](value))

    return document


def _value_to_str(value: Any) -> str:
    if value is None:
        return "n/a"
    elif isinstance(value, vp.Color):
        value_str = value.as_hex()
        value_type = "color"
    else:
        value_str = str(value)
        value_type = type(value).__name__
    return f"({value_type}) {value_str}"


@cli.command(group="Metadata")
@click.option("--global", "-g", "global_flag", is_flag=True, help="Global mode.")
@click.option("-l", "--layer", type=LayerType(accept_multiple=True), help="Target layer(s).")
@global_processor
def proplist(document: vp.Document, global_flag: bool, layer: int | list[int] | None):
    """Print a list the existing global or layer properties and their values.

    Either the `--global` or `--layer` option must be used to specify whether global or
    layer properties should be listed. When using `--layer`, either a single layer ID, a
    coma-separated list of layer ID, or `all` may be used.

    Examples:

        Print a list of global properties:

            vpype pagesize a4 proplist -g
    """
    global_flag, layer = _check_scope(global_flag, layer)

    if global_flag:
        print(f"listing {len(document.metadata)} global properties")
        for prop in sorted(document.metadata.keys()):
            print(f"  {prop}: {_value_to_str(document.property(prop))}")
    else:
        for lid in multiple_to_layer_ids(layer, document):
            lc = document.layers[lid]
            print(f"listing {len(lc.metadata)} properties for layer {lid}")
            for prop in sorted(lc.metadata.keys()):
                print(f"  {prop}: {_value_to_str(lc.property(prop))}")

    return document


@cli.command(group="Metadata")
@click.argument("prop", type=TextType())
@click.option("--global", "-g", "global_flag", is_flag=True, help="Global mode.")
@click.option("-l", "--layer", type=LayerType(accept_multiple=True), help="Target layer(s).")
@global_processor
def propget(
    document: vp.Document, global_flag: bool, layer: int | list[int] | None, prop: str
):
    """Print the value of a global or layer property.

    Either the `--global` or `--layer` option must be used to specify whether a global or
    layer property should be displayed. When using `--layer`, either a single layer ID, a
    coma-separated list of layer ID, or `all` may be used.

    Examples:

        Print the value of property `vp_color` for all layers:

            vpype [...] pens cmyk propget --layer all vp_color [...]
    """
    global_flag, layer = _check_scope(global_flag, layer)

    if global_flag:
        print(f"global property {prop}: {_value_to_str(document.property(prop))}")
    else:
        for lid in multiple_to_layer_ids(layer, document):
            print(
                f"layer {lid} property {prop}: "
                f"{_value_to_str(document.layers[lid].property(prop))}"
            )

    return document


@cli.command(group="Metadata")
@click.argument("prop", type=TextType())
@click.option("--global", "-g", "global_flag", is_flag=True, help="Global mode.")
@click.option("-l", "--layer", type=LayerType(accept_multiple=True), help="Target layer(s).")
@global_processor
def propdel(
    document: vp.Document, global_flag: bool, layer: int | list[int] | None, prop: str
):
    """Remove a global or layer property.

    Either the `--global` or `--layer` option must be used to specify whether a global or
    layer property should be removed. When using `--layer`, either a single layer ID, a
    coma-separated list of layer ID, or `all` may be used.

    Examples:

        Remove a property from a layer:

            vpype [...] pens cmyk propdel --layer 1 vp_name [...]
    """
    global_flag, layer = _check_scope(global_flag, layer)

    if global_flag:
        document.set_property(prop, None)
    else:
        for lid in multiple_to_layer_ids(layer, document):
            document.layers[lid].set_property(prop, None)

    return document


@cli.command(group="Metadata")
@click.option("--global", "-g", "global_flag", is_flag=True, help="Global mode.")
@click.option("-l", "--layer", type=LayerType(accept_multiple=True), help="Target layer(s).")
@global_processor
def propclear(document: vp.Document, global_flag: bool, layer: int | list[int] | None):
    """Remove all global or layer properties.

    Either the `--global` or `--layer` option must be used to specify whether global or
    layer properties should be cleared. When using `--layer`, either a single layer ID, a
    coma-separated list of layer ID, or `all` may be used.

    Examples:

        Remove all global properties:

            vpype [...] propclear --global [...]

        Remove all properties from layer 1 and 2:

            vpype [...] propclear --layer 1,2 [...]
    """
    global_flag, layer = _check_scope(global_flag, layer)

    if global_flag:
        document.clear_metadata()
    else:
        for lid in multiple_to_layer_ids(layer, document):
            document.layers[lid].clear_metadata()

    return document


@cli.command(group="Metadata")
@click.argument("pen_width", type=LengthType(), metavar="WIDTH")
@layer_processor
def penwidth(layer: vp.LineCollection, pen_width: float) -> vp.LineCollection:
    """Set the pen width for one or more layers.

    By default, this commands sets the pen width for all layers. Use the `--layer` option to
    set the pen width of one (or more) specific layer(s).

    Examples:

        Set the pen width for all layers:

            $ vpype [...] penwidth 0.15mm [...]

        Set the pen width for a specific layer:

            $ vpype [...] penwidth --layer 2 0.15mm [...]
    """
    layer.set_property(vp.METADATA_FIELD_PEN_WIDTH, pen_width)
    return layer


# noinspection PyShadowingNames
@cli.command(group="Metadata")
@click.argument("color", type=TextType())
@layer_processor
def color(layer: vp.LineCollection, color: str) -> vp.LineCollection:
    """Set the color for one or more layers.

    Any SVG-compatible string may be used for VALUE, including 16-bit RGB (#ff0000),
    16-bit RGBA (#ff0000ff), 8-bit variants (#f00 or #f00f), or color names (red).

    By default, this commands sets the color for all layers. Use the `--layer` option to set
    the color of one (or more) specific layer(s).

    Note: the opacity of a layer may be set without changing the base color with the `alpha`
    command.

    Examples:

        Set the color for all layers:

            $ vpype [...] color red [...]

        Set the color for a specific layer:

            $ vpype [...] color --layer 2 #0f0 [...]
    """

    layer.set_property(vp.METADATA_FIELD_COLOR, vp.Color(color))
    return layer


# noinspection PyShadowingNames
@cli.command(group="Metadata")
@click.argument("alpha", metavar="ALPHA", type=FloatRangeType(min=0.0, max=1.0, clamp=True))
@layer_processor
def alpha(layer: vp.LineCollection, alpha: float) -> vp.LineCollection:
    """Set the opacity of one or more layers.

    Assign the opacity ALPHA to the target layer(s) without changing its based color. If the
    base color is undefined, it is set to black.

    By default, this commands operate on all layers. Use the `--layer` option to specify one or
    more target layer(s)

    Examples:

        Set all layer to 50% red:

            $ vpype [...] color red alpha .5 [...]

        Set the opacity for a specific layer:

            $ vpype [...] alpha --layer 2 .5 [...]
    """

    if (color := layer.property(vp.METADATA_FIELD_COLOR)) is None:
        color = vp.Color("black")
    layer.set_property(
        vp.METADATA_FIELD_COLOR, vp.Color(color.red, color.green, color.blue, int(alpha * 255))
    )
    return layer


# noinspection PyShadowingNames
@cli.command(group="Metadata")
@click.argument("name", type=TextType())
@layer_processor
def name(layer: vp.LineCollection, name: str) -> vp.LineCollection:
    """Set the name for one or more layers.

    By default, this commands sets the name for all layers. Use the `--layer` option to set
    the name of one (or more) specific layer(s).

    Examples:

        Set the name for a specific layer:

            $ vpype [...] name --layer 4 black [...]
    """

    layer.set_property(vp.METADATA_FIELD_NAME, name)
    return layer


PENS_HELP_STRING = f"""Apply a pen configuration.

This command applies given names, pen colors and/or pen widths to one or more layers, as
defined by the pen configuration CONF. This pen configuration just be defined in either the
bundled or a user-provided config file (such as ~/.vpype.toml).

The following pen configurations are defined in the default config and the ~/.vpype.toml file:

    {", ".join(vp.config_manager.config.get("pen_config", {}).keys()) or "n/a"}

For example, the bundled pen configuration `cmyk` applies `cyan`,`magenta`, `yellow`, resp.
`black` to the name and color of layers 1 to 4 while leaving pen widths unchanged.

In details, for each of the layers defined in a pen configuration, this command performs the
following tasks:
- create the corresponding layer if it does not exist
- set the name of the layer as specified by the pen configuration (if specified)
- set the color of the layer as specified by the pen configuration (if specified)
- set the pen width of the layer as specified by the pen configuration (if specified)

Existing layers whose ID are not included in the pen configuration are not affected by this
command. Check the documentation for more information on creating custom pen configurations.
"""


@cli.command(group="Metadata", help=PENS_HELP_STRING)
@click.argument("pen_config", metavar="CONF", type=TextType())
@global_processor
def pens(document: vp.Document, pen_config: str) -> vp.Document:

    # the CONF parameter must be checked explicitly (instead of using click.Choice()) because
    # additional config file (potentially containing pen configs) may be added at runtime
    config = vp.config_manager.config.get("pen_config", {})

    if pen_config not in config:
        click.secho(
            f"pens: pen configuration '{pen_config}' not found, no pen configuration applied",
            fg="red",
        )
        return document

    for layer_data in config[pen_config]["layers"]:
        lid = layer_data["layer_id"]
        document.add([], lid)
        if "name" in layer_data:
            document.layers[lid].set_property(vp.METADATA_FIELD_NAME, layer_data["name"])
        if "color" in layer_data:
            document.layers[lid].set_property(vp.METADATA_FIELD_COLOR, layer_data["color"])
        if "pen_width" in layer_data:
            document.layers[lid].set_property(
                vp.METADATA_FIELD_PEN_WIDTH, vp.convert_length(layer_data["pen_width"])
            )

    return document



from __future__ import annotations

import logging

import click

import vpype as vp

from .cli import cli
from .decorators import global_processor, pass_state
from .state import State


@cli.command("eval", group="Expressions")
@click.argument("expr", type=str)
@global_processor
@pass_state
def eval_cmd(state: State, document: vp.Document, expr: str):
    """Evaluate an expression.

    This command is a placeholder whose only purpose is to evaluate EXPR. It has no effect on
    the geometries, nor their properties. It is typically used at the beginning of a pipeline
    to initialise expression variables used later on.

    EXPR is interpreted as an expression in its entirety. As such, the enclosing `%` expression
    markers may be omitted.

    Example:

        Crop the geometry to a 1-cm margin.

    \b
            $ vpype read input.svg eval "m = 1*cm; w,h = prop.vp_page_size" \\
                crop %m% %m% %w-2*m% %h-2*m% write output.svg
    """

    if not expr.startswith("%"):
        expr = "%" + expr
    if not expr.endswith("%"):
        expr += "%"
    res = state.substitute(expr)
    logging.debug(f"eval: expression evaluated to {res}")

    return document



from __future__ import annotations

import glob
import os
import pathlib
from typing import Any, Iterable

import click

import vpype as vp

from .cli import ProcessorType, cli, execute_processors
from .decorators import block_processor
from .state import State
from .types import IntegerType, LengthType, TextType

__all__ = ("grid", "repeat", "forfile", "forlayer")


@cli.command(group="Block processors")
@click.argument("number", nargs=2, default=(2, 2), type=IntegerType(), metavar="NX NY")
@click.option(
    "-o",
    "--offset",
    nargs=2,
    default=("10mm", "10mm"),
    type=LengthType(),
    metavar="DX DY",
    help="Offset between columns and rows.",
)
@click.option(
    "-k",
    "--keep-page-size",
    "keep_page_size",
    is_flag=True,
    help="Do not change the page size.",
)
@block_processor
def grid(
    state: State,
    processors: Iterable[ProcessorType],
    number: tuple[int, int],
    offset: tuple[float, float],
    keep_page_size: bool,
) -> None:
    """Creates a NX by NY grid of geometry

    The number of column and row must always be specified. By default, 10mm offsets are used
    in both directions. Use the `--offset DX DY` option to override these values.

    The nested commands are exposed to a pipeline which does not contain any geometry but
    retains the layer structure and metadata. The properties created and modified by the
    nested commands are applied on the pipeline. However, the properties deleted by the nested
    commands are not deleted from the outer pipeline.

    By default, the `grid` block changes the current page size according to its geometry (e.g.
    it sets the page size to (NxDX, MxDY). This can be prevented using the `--keep-page-size`
    option.

    The following variables are set by `grid` and available for expressions:

    \b
        _nx: number of columns (NX)
        _ny: number of rows (NY)
        _n: total number of cells (NX*NY)
        _x: current column (0 to NX-1)
        _y: current row (0 to NY-1)
        _i: current cell (0 to _n-1)

    Examples:

        Create a grid of random line patches:

            $ vpype begin grid 3 4 random end show

        Create a grid of circles, each on a different layer:

            $ vpype begin grid -o 3cm 3cm 2 3 circle --layer new 0 0 1cm end show
    """

    nx, ny = number
    orig_doc = state.document

    for j in range(ny):
        for i in range(nx):
            state.document = orig_doc.clone(keep_layers=True)
            variables = {
                "_nx": nx,
                "_ny": ny,
                "_n": nx * ny,
                "_x": i,
                "_y": j,
                "_i": i + j * nx,
            }
            with state.expression_variables(variables):
                execute_processors(processors, state)
            state.document.translate(offset[0] * i, offset[1] * j)
            orig_doc.extend(state.document)

    state.document = orig_doc

    if nx > 0 and ny > 0 and not keep_page_size:
        state.document.page_size = (nx * offset[0], ny * offset[1])


@cli.command(group="Block processors")
@click.argument("number", type=IntegerType(), metavar="N")
@block_processor
def repeat(state: State, processors: Iterable[ProcessorType], number: int) -> None:
    """Repeat geometries N times.

    Repeats the enclosed command N times, stacking their output on top of each other.

    The nested commands are exposed to a pipeline which does not contain any geometry but
    retains the layer structure and metadata. The properties created and modified by the
    nested commands are applied on the pipeline. However, the properties deleted by the nested
    commands are not deleted from the outer pipeline.

    The following variables are set by `repeat` and available for expressions:

    \b
        _n: number of repetitions (N)
        _i: counter (0 to N-1)

    Examples:

        Create a patch of random lines of 3 different colors:

            $ vpype begin repeat 3 random --layer %_i+1% end show
    """

    orig_doc = state.document
    for i in range(number):
        state.document = orig_doc.clone(keep_layers=True)
        variables = {"_n": number, "_i": i}
        with state.expression_variables(variables):
            execute_processors(processors, state)
        orig_doc.extend(state.document)
    state.document = orig_doc


@cli.command(group="Block processors")
@click.argument("files", type=TextType(), metavar="FILES")
@block_processor
def forfile(state: State, processors: Iterable[ProcessorType], files: str) -> None:
    """Iterate over a file list.

    The `forfile` block processor expends the FILES pattern into a file list like a shell
    would. In particular, wildcards (`*` and `**`) are expended, environmental variables are
    substituted (`$ENVVAR`), and the `~` expended to the user directory. It then executes the
    nested commands once for each file.

    The nested commands are exposed to a pipeline which does not contain any geometry but
    retains the layer structure and metadata. The properties created and modified by the
    nested commands are applied on the pipeline. However, the properties deleted by the nested
    commands are not deleted from the outer pipeline.

    The following variables are set by `forfile` and available for expressions:

    \b
        _path (pathlib.Path): file path
        _name (str): file name, same as _path.name (e.g. 'input.svg')
        _parent (pathlib.Path): parent directory, same as _path.parent
        _ext (str): file extension, same as _path.suffix (e.g. '.svg')
        _stem (str): file name without extension, same as _path.stem (e.g. 'input')
        _n (int): total number of files
        _i (int): counter (0 to _n-1)

    Example:

        Process all SVGs in the current directory:

    \b
            $ vpype begin forfile \\*.svg read %_path% linemerge linesort \\
                write "optimized/%basename(_path)%" end
    """

    file_list = sorted(
        glob.glob(os.path.expandvars(os.path.expanduser(files))), key=os.path.basename
    )
    orig_document = state.document

    for i, file in enumerate(file_list):
        state.document = orig_document.clone(keep_layers=True)
        path = pathlib.Path(file)
        variables = {
            "_path": path,
            "_name": path.name,
            "_parent": path.parent,
            "_ext": path.suffix,
            "_stem": path.stem,
            "_i": i,
            "_n": len(files),
        }
        with state.expression_variables(variables):
            execute_processors(processors, state)
        orig_document.extend(state.document)

    state.document = orig_document


class _MetadataProxy:
    def __init__(self, metadata: dict[str, Any]):
        self._metadata = metadata

    def __getattr__(self, name):
        if name in self._metadata:
            return self._metadata[name]
        else:
            raise AttributeError(f"property '{name}' not found")

    def __setattr__(self, name, value):
        if name == "_metadata":
            super().__setattr__(name, value)
        else:
            self._metadata[name] = value

    def __getitem__(self, name):
        return self._metadata[name]

    def __setitem__(self, name, value):
        self._metadata[name] = value


@cli.command(group="Block processors")
@block_processor
def forlayer(state: State, processors: Iterable[ProcessorType]) -> None:
    """Iterate over each layer.

    This block processor execute the nested pipeline once per layer. The nested pipeline is
    exclusively exposed to the current layer. In addition, if the nested commands create any
    other layer, they are merged as well into the outer pipeline.

    The following variables are set by `forlayer` and available for expressions:

    \b
        _lid (int): the current layer ID
        _name (str): the name of the current layer
        _color (Color): the color of the current layer
        _pen_width (float): the pen width of the current layer
        _prop: properties of the current layer (accessible by item and/or attribute)
        _i (int): counter (0 to _n-1)
        _n (int): number of layers

    Example:

        Export one file per layer:

            vpype read input.svg forlayer write output_%_name%.svg end
    """

    orig_doc = state.document
    new_doc: vp.Document = orig_doc.clone()

    lids = list(orig_doc.layers)
    for i, lid in enumerate(lids):
        state.document = orig_doc.clone()
        state.document.add(orig_doc.pop(lid), lid, with_metadata=True)

        variables = {
            "_lid": lid,
            "_i": i,
            "_n": len(lids),
            "_name": state.document.layers[lid].property(vp.METADATA_FIELD_NAME) or "",
            "_color": state.document.layers[lid].property(vp.METADATA_FIELD_COLOR),
            "_pen_width": state.document.layers[lid].property(vp.METADATA_FIELD_PEN_WIDTH),
            "_prop": _MetadataProxy(state.document.layers[lid].metadata),
        }
        with state.expression_variables(variables):
            execute_processors(processors, state)
        new_doc.extend(state.document)

    state.document = new_doc



"""
Generic viewer
"""
from __future__ import annotations

import enum
from collections import defaultdict
from typing import Callable

import moderngl as mgl
import numpy as np

import vpype as vp

from ._painters import (
    LineCollectionFastColorfulPainter,
    LineCollectionFastPainter,
    LineCollectionPenUpPainter,
    LineCollectionPointsPainter,
    LineCollectionPreviewPainter,
    Painter,
    PaperBoundsPainter,
    RulersPainter,
)
from ._scales import DEFAULT_SCALE_SPEC, SCALES_MAP, ScaleSpec, UnitType
from ._utils import ColorType, orthogonal_projection_matrix

__all__ = ["DEFAULT_PEN_WIDTH", "DEFAULT_PEN_OPACITY", "ViewMode", "Engine"]

DEFAULT_PEN_WIDTH = 1.1  # about 0.3mm
DEFAULT_PEN_OPACITY = 0.8


class ViewMode(enum.Enum):
    """View mode enum."""

    NONE = enum.auto()  # for debug purposes
    OUTLINE = enum.auto()
    OUTLINE_COLORFUL = enum.auto()
    PREVIEW = enum.auto()


class Engine:
    """OpenGL-based rendering engine for :class:`vpype.Document`."""

    def __init__(
        self,
        view_mode: ViewMode = ViewMode.OUTLINE,
        show_pen_up: bool = False,
        show_points: bool = False,
        show_rulers: bool = True,
        default_pen_width: float = DEFAULT_PEN_WIDTH,
        default_pen_opacity: float = DEFAULT_PEN_OPACITY,
        override_pen_width: bool = False,
        override_pen_opacity: bool = False,
        render_cb: Callable[[], None] = lambda: None,
    ):
        """Constructor.

        Args:
            view_mode: the view mode to use
            show_pen_up: render pen-up trajectories if True
            show_points: render points if True
            show_rulers: display the rulers
            default_pen_width: pen width (preview only)
            default_pen_opacity: pen opacity (preview only)
            override_pen_width: disregard pen width from properties and use
                ``default_pen_width`` instead (preview only)
            override_pen_opacity: disregard pen opacity from properties and use
                ``default_pen_opacity`` instead (preview only)
            render_cb: callback that will be called when rendering is required
        """

        # params
        self._debug = False
        self._view_mode = view_mode
        self._show_pen_up = show_pen_up
        self._show_points = show_points
        self._show_rulers = show_rulers
        self._default_pen_width = default_pen_width
        self._default_pen_opacity = default_pen_opacity
        self._override_pen_width = override_pen_width
        self._override_pen_opacity = override_pen_opacity
        self._render_cb = render_cb
        self._unit_type = UnitType.METRIC
        self._scale_threshold = 100.0  # min size for a major tick in rulers

        # state
        self._pixel_factor = 1.0
        self._ctx: mgl.Context | None = None
        self._viewport_width = 100
        self._viewport_height = 100
        self._scale = 1.0  # one pixel of page equal one pixel of view port
        self._origin = (0.0, 0.0)  # top-left of page aligned with top-left of view port
        self._document: vp.Document | None = None
        self._rebuild_needed = True
        self._scale_spec = DEFAULT_SCALE_SPEC

        # painters
        self._layer_visibility: dict[int, bool] = defaultdict(lambda: True)
        self._layer_painters: dict[int, list[Painter]] = defaultdict(list)
        self._paper_bounds_painter: PaperBoundsPainter | None = None
        self._rulers_painter: RulersPainter | None = None

        self._fit_to_viewport_flag = True

    def post_init(self, ctx: mgl.Context, width: int = 100, height: int = 100):
        """Post-init configuration to provide a GL context."""

        self._ctx = ctx
        self._ctx.enable_only(mgl.BLEND | mgl.PROGRAM_POINT_SIZE)

        self._rulers_painter = RulersPainter(self._ctx)

        self.resize(width, height)

    # =========================================================================================
    # Properties

    @property
    def document(self) -> vp.Document | None:
        """:class:`vpype.Document` being displayed."""
        return self._document

    @document.setter
    def document(self, document: vp.Document | None) -> None:
        self._document = document
        self._layer_visibility.clear()

        # keep fitted to the viewport
        if self._fit_to_viewport_flag:
            self.fit_to_viewport()

        self._update()

    @property
    def scale(self) -> float:
        """Current scale (ratio screen pixel per document pixel)"""
        return self._scale

    @scale.setter
    def scale(self, scale: float) -> None:
        self._scale = scale
        self._fit_to_viewport_flag = False
        self._update(False)

    @property
    def origin(self) -> tuple[float, float]:
        """Current origin (document coordinates corresponding to the display window's top-left
        corner."""
        return self._origin

    @origin.setter
    def origin(self, origin: tuple[float, float]):
        self._origin = origin
        self._fit_to_viewport_flag = False
        self._update(False)

    @property
    def width(self) -> float:
        return self._viewport_width

    @property
    def height(self) -> float:
        return self._viewport_height

    @property
    def view_mode(self) -> ViewMode:
        """Current view mode."""
        return self._view_mode

    @view_mode.setter
    def view_mode(self, view_mode: ViewMode) -> None:
        self._view_mode = view_mode
        self._update()

    @property
    def show_pen_up(self) -> bool:
        """True if pen-up trajectories are rendered."""
        return self._show_pen_up

    @show_pen_up.setter
    def show_pen_up(self, show_pen_up: bool):
        self._show_pen_up = show_pen_up
        self._update()

    @property
    def show_points(self) -> bool:
        """True if points are rendered."""
        return self._show_points

    @show_points.setter
    def show_points(self, show_points: bool):
        self._show_points = show_points
        self._update()

    @property
    def show_rulers(self) -> bool:
        return self._show_rulers

    @show_rulers.setter
    def show_rulers(self, show_rulers: bool) -> None:
        self._show_rulers = show_rulers
        if self._fit_to_viewport_flag:
            self.fit_to_viewport()
        self._update(False)

    @property
    def pixel_factor(self) -> float:
        return self._pixel_factor

    @pixel_factor.setter
    def pixel_factor(self, pixel_factor: float) -> None:
        self._pixel_factor = pixel_factor
        if self._fit_to_viewport_flag:
            self.fit_to_viewport()
        self._update(False)

    @property
    def scale_spec(self) -> ScaleSpec:
        return self._scale_spec

    @property
    def unit_type(self) -> UnitType:
        return self._unit_type

    @unit_type.setter
    def unit_type(self, unit_type: UnitType) -> None:
        self._unit_type = unit_type
        self._update(False)

    @property
    def default_pen_width(self) -> float:
        """Pen width used for rendering (preview only)."""
        return self._default_pen_width

    @default_pen_width.setter
    def default_pen_width(self, pen_width: float):
        self._default_pen_width = pen_width
        self._update()

    @property
    def default_pen_opacity(self) -> float:
        """Pen opacity used for rendering (preview only)."""
        return self._default_pen_opacity

    @default_pen_opacity.setter
    def default_pen_opacity(self, pen_opacity: float):
        self._default_pen_opacity = pen_opacity
        self._update()

    @property
    def override_pen_width(self) -> bool:
        """Disregard pen opacity from properties (preview only)."""
        return self._override_pen_width

    @override_pen_width.setter
    def override_pen_width(self, override: bool):
        self._override_pen_width = override
        self._update()

    @property
    def override_pen_opacity(self) -> bool:
        """Disregard pen opacity from properties (preview only)."""
        return self._override_pen_opacity

    @override_pen_opacity.setter
    def override_pen_opacity(self, override: bool):
        self._override_pen_opacity = override
        self._update()

    @property
    def debug(self) -> bool:
        """Debug mode for display."""
        return self._debug

    @debug.setter
    def debug(self, debug: bool):
        self._debug = debug
        self._update(False)

    def layer_visible(self, layer_id: int) -> bool:
        """True if the corresponding layer is currently visible.

        Args:
            layer_id: layer to check
        """
        return self._layer_visibility[layer_id]

    def toggle_layer_visibility(self, layer_id: int) -> None:
        """Toggles the visibility of the provided layer.

        Args:
            layer_id: layer to toggle
        """
        self._layer_visibility[layer_id] = not self._layer_visibility[layer_id]
        self._update(False)

    # =========================================================================================
    # Geometry

    def fit_to_viewport(self):
        """Fit the current document in the viewport, allowing for a 2.5% margin on either
        sides."""
        if self._document is None:
            return

        if self._document.page_size is not None and 0 not in self._document.page_size:
            x1, y1 = 0.0, 0.0
            x2, y2 = self._document.page_size
        else:
            bounds = self._document.bounds()
            if bounds is None:
                return
            x1, y1, x2, y2 = bounds

        w = x2 - x1
        h = y2 - y1

        if self.show_rulers and self._rulers_painter is not None:
            ruler_thickness = self._rulers_painter.thickness * self.pixel_factor
        else:
            ruler_thickness = 0.0
        viewport_width = self._viewport_width - ruler_thickness
        viewport_height = self._viewport_height - ruler_thickness
        self._scale = 0.95 * min(viewport_width / w, viewport_height / h)
        self._origin = [
            x1 - (viewport_width / self._scale - w) / 2 - ruler_thickness / self.scale,
            y1 - (viewport_height / self._scale - h) / 2 - ruler_thickness / self.scale,
        ]

        self._fit_to_viewport_flag = True

        self._update(False)

    def viewport_to_model(self, x: float, y: float) -> tuple[float, float]:
        """Converts viewport coordinates to model coordinates in pixels."""
        return x / self._scale + self._origin[0], y / self._scale + self._origin[1]

    def resize(self, width: int, height: int) -> None:
        """Resizes the viewport.
        Args:
            width: new viewport width
            height: new viewport height
        """
        self._viewport_width = max(width, 1)
        self._viewport_height = max(height, 1)

        if self._fit_to_viewport_flag:
            self.fit_to_viewport()

    def _get_projection(self) -> np.ndarray:
        proj = orthogonal_projection_matrix(
            self._origin[0],
            self._origin[0] + self._viewport_width / self._scale,
            self._origin[1] + self._viewport_height / self._scale,
            self._origin[1],
            -1,
            1,
            dtype="f4",
        )

        return proj

    def pan(self, dx: float, dy: float) -> None:
        """Pan the viewport.

        Args:
            dx: horizontal distance to pan
            dy: vertical distance to pan
        """

        if dx != 0 or dy != 0:
            self._fit_to_viewport_flag = False

        self._origin = (self._origin[0] - dx / self._scale, self._origin[1] - dy / self._scale)
        self._update(False)

    def zoom(self, delta_zoom: float, mouse_x: float, mouse_y: float) -> None:
        """Zoom the viewport.

        Coordinates must be provided to specify the point about which to zoom. This typically
        should be the mouse position when used in GUI.

        Args:
            delta_zoom: zoom adjustment value (the scale is multiplied by ``(1 + delta_zoom)``
            mouse_x: mouse X coordinate
            mouse_y: mouse Y coordinate
        """
        if delta_zoom != 0:
            self._fit_to_viewport_flag = False

        new_scale = self._scale * (1 + delta_zoom)
        new_scale = max(min(new_scale, 100000), 0.05)  # clamp to reasonable values

        dz = 1 / self._scale - 1 / new_scale
        self._origin = (self._origin[0] + mouse_x * dz, self._origin[1] + mouse_y * dz)
        self._scale = new_scale

        self._update(False)

    # =========================================================================================
    # Painters

    def render(self):
        """Execute a render."""
        if self._ctx is None:
            return

        if self._rebuild_needed:
            self._rebuild()

        # update scale spec
        scales = SCALES_MAP[self.unit_type]
        threshold = self._scale_threshold * self.pixel_factor
        for spec in scales:
            if spec.scale_px * self.scale < threshold:
                break
        else:
            spec = scales[-1]
        self._scale_spec = spec

        self._ctx.clear(0.95, 0.95, 0.95, 1)
        proj = self._get_projection()

        if self._paper_bounds_painter:
            self._paper_bounds_painter.render(self, proj)

        for layer_id in sorted(self._layer_painters):
            if self._layer_visibility[layer_id]:
                for painter in self._layer_painters[layer_id]:
                    painter.render(self, proj)

        if self._rulers_painter and self._show_rulers:
            self._rulers_painter.render(self, proj)

    def _update(self, rebuild=True):
        """"""
        if rebuild:
            self._rebuild_needed = True
        self._render_cb()

    def _rebuild(self):
        self._layer_painters.clear()
        self._paper_bounds_painter = None

        if self._ctx is None:
            return

        if self._document is not None:
            color_index = 0
            for layer_id in sorted(self._document.layers):
                default_color = vp.METADATA_DEFAULT_COLOR_SCHEME[
                    color_index % len(vp.METADATA_DEFAULT_COLOR_SCHEME)
                ]
                color_index += 1
                layer_color: ColorType = vp.Color(
                    default_color.red,
                    default_color.green,
                    default_color.blue,
                    int(self.default_pen_opacity * 255),
                ).as_floats()
                layer_color_no_transparency = (
                    layer_color[0],
                    layer_color[1],
                    layer_color[2],
                    1.0,
                )

                lc = self._document.layers[layer_id]
                if lc.is_empty():
                    continue

                if vp.METADATA_FIELD_COLOR in lc.metadata:
                    color = lc.metadata[vp.METADATA_FIELD_COLOR]
                    layer_color = color.as_floats()
                    if self.override_pen_opacity:
                        layer_color = layer_color[0:3] + (self.default_pen_opacity,)

                if self.view_mode == ViewMode.OUTLINE:
                    self._layer_painters[layer_id].append(
                        LineCollectionFastPainter(
                            self._ctx, lc=lc, color=layer_color_no_transparency
                        )
                    )
                elif self.view_mode == ViewMode.OUTLINE_COLORFUL:
                    self._layer_painters[layer_id].append(
                        LineCollectionFastColorfulPainter(
                            self._ctx, lc=lc, show_points=self._show_points
                        )
                    )
                elif self.view_mode == ViewMode.PREVIEW:
                    if self.override_pen_width:
                        pen_width = self._default_pen_width
                    else:
                        pen_width = lc.metadata.get(
                            vp.METADATA_FIELD_PEN_WIDTH, self._default_pen_width
                        )
                    self._layer_painters[layer_id].append(
                        LineCollectionPreviewPainter(
                            self._ctx, lc=lc, pen_width=pen_width, color=layer_color
                        )
                    )

                if self.show_pen_up:
                    self._layer_painters[layer_id].append(
                        LineCollectionPenUpPainter(self._ctx, lc=lc)
                    )
                if self.show_points and self.view_mode != ViewMode.OUTLINE_COLORFUL:
                    self._layer_painters[layer_id].append(
                        LineCollectionPointsPainter(
                            self._ctx, lc=lc, color=layer_color_no_transparency
                        )
                    )

            page_size = self._document.page_size
            if page_size is not None:
                self._paper_bounds_painter = PaperBoundsPainter(
                    self._ctx, self._document.page_size
                )

        self._rebuild_needed = False



from __future__ import annotations

import enum
from dataclasses import dataclass

MM_TO_PIXELS = 96.0 / 25.4
FT_TO_PIXELS = 12 * 96.0
IN_TO_PIXELS = 96.0


class UnitType(enum.Enum):
    """Type of unit to use for display purposes."""

    METRIC = enum.auto()
    IMPERIAL = enum.auto()
    PIXELS = enum.auto()


@dataclass
class ScaleSpec:
    scale: int  # in display unit
    divisions: tuple[int, int, int]  # total sub-division, mid tick, small tick
    to_px: float  # conversion to px
    unit: str  # unit to display

    @property
    def scale_px(self) -> float:
        return self.scale * self.to_px


DEFAULT_SCALE_SPEC = ScaleSpec(1, (10, 5, 1), 1.0, "px")

# scale, division, mid, minor
PIXEL_SCALES = (
    ScaleSpec(5000, (10, 2, 1), 1.0, "px"),
    ScaleSpec(2000, (10, 5, 1), 1.0, "px"),
    ScaleSpec(1000, (10, 5, 1), 1.0, "px"),
    ScaleSpec(500, (10, 2, 1), 1.0, "px"),
    ScaleSpec(200, (10, 5, 1), 1.0, "px"),
    ScaleSpec(100, (10, 5, 1), 1.0, "px"),
    ScaleSpec(50, (10, 2, 1), 1.0, "px"),
    ScaleSpec(20, (10, 5, 1), 1.0, "px"),
    ScaleSpec(10, (10, 5, 1), 1.0, "px"),
    ScaleSpec(5, (10, 2, 1), 1.0, "px"),
    ScaleSpec(2, (10, 5, 1), 1.0, "px"),
    ScaleSpec(1, (10, 5, 1), 1.0, "px"),
)

METRIC_SCALES = (
    ScaleSpec(10, (10, 5, 1), 1000 * MM_TO_PIXELS, "m"),
    ScaleSpec(5, (10, 2, 1), 1000 * MM_TO_PIXELS, "m"),
    ScaleSpec(2, (10, 5, 1), 1000 * MM_TO_PIXELS, "m"),
    ScaleSpec(1, (10, 5, 1), 1000 * MM_TO_PIXELS, "m"),
    ScaleSpec(50, (10, 2, 1), 10 * MM_TO_PIXELS, "cm"),
    ScaleSpec(20, (10, 5, 1), 10 * MM_TO_PIXELS, "cm"),
    ScaleSpec(10, (10, 5, 1), 10 * MM_TO_PIXELS, "cm"),
    ScaleSpec(5, (10, 2, 1), 10 * MM_TO_PIXELS, "cm"),
    ScaleSpec(2, (10, 5, 1), 10 * MM_TO_PIXELS, "cm"),
    ScaleSpec(1, (10, 5, 1), 10 * MM_TO_PIXELS, "cm"),
    ScaleSpec(5, (10, 2, 1), MM_TO_PIXELS, "mm"),
    ScaleSpec(2, (10, 5, 1), MM_TO_PIXELS, "mm"),
    ScaleSpec(1, (10, 5, 1), MM_TO_PIXELS, "mm"),
)

IMPERIAL_SCALES = (
    ScaleSpec(100, (10, 5, 1), IN_TO_PIXELS, "in"),
    ScaleSpec(50, (10, 2, 1), IN_TO_PIXELS, "in"),
    ScaleSpec(20, (10, 5, 1), IN_TO_PIXELS, "in"),
    ScaleSpec(10, (10, 5, 1), IN_TO_PIXELS, "in"),
    ScaleSpec(5, (10, 2, 1), IN_TO_PIXELS, "in"),
    ScaleSpec(2, (16, 8, 4), IN_TO_PIXELS, "in"),
    ScaleSpec(1, (16, 8, 4), IN_TO_PIXELS, "in"),
)


SCALES_MAP = {
    UnitType.METRIC: METRIC_SCALES,
    UnitType.IMPERIAL: IMPERIAL_SCALES,
    UnitType.PIXELS: PIXEL_SCALES,
}



from __future__ import annotations

import glob
import os
import pathlib
import sys
from typing import TYPE_CHECKING, Any, Callable, Iterable

import asteval

import vpype as vp

if TYPE_CHECKING:  # pragma no cover
    from .state import State


class PropertySubstitutionError(Exception):
    """Error while performing property substitution."""


class ExpressionSubstitutionError(Exception):
    """Error while performing expression substitution."""


class _PropertyProxy:
    """Helper proxy class to provide access to global and current layer properties.

    Used for :ref:`property substitution <fundamentals_property_substitution>`
    using :meth:`str.format_map` and for expression evaluation.

    Either or both global and current layer properties may be configured. If both are enabled,
    current layer's properties take precedence over global properties.

    Args:
        state: :class:`State` instance
        global_prop: access global properties
        layer_prop: access current layer properties
    """

    def __init__(self, state: State, global_prop: bool = True, layer_prop: bool = True):
        self._state = state
        self._global_prop = global_prop
        self._layer_prop = layer_prop

    def _get_prop(self, name) -> Any | None:
        if self._state.document is None:
            return None

        if (
            self._layer_prop
            and self._state.current_layer_id is not None
            and self._state.document.exists(self._state.current_layer_id)
            and self._state.document.layers[self._state.current_layer_id].property_exists(name)
        ):
            return self._state.document.layers[self._state.current_layer_id].property(name)
        elif self._global_prop and self._state.document.property_exists(name):
            return self._state.document.property(name)
        else:
            return None

    def _set_prop(self, name, value) -> None:
        if self._state.document is None or self._layer_prop == self._global_prop:
            raise TypeError("cannot set property value (read-only)")

        if self._layer_prop:
            if self._state.current_layer_id is None or not self._state.document.exists(
                self._state.current_layer_id
            ):
                raise TypeError("cannot set property value (unknown current layer)")
            else:
                self._state.document.layers[self._state.current_layer_id].set_property(
                    name, value
                )
        elif self._global_prop:
            self._state.document.set_property(name, value)

    def keys(self):
        k = set()
        if self._global_prop and self._state.document is not None:
            k |= self._state.document.metadata.keys()
        if (
            self._layer_prop
            and self._state.current_layer_id is not None
            and self._state.document.exists(self._state.current_layer_id)
        ):
            k |= self._state.document.layers[self._state.current_layer_id].metadata.keys()
        return k

    def __getitem__(self, name):
        prop = self._get_prop(name)
        if prop is not None:
            return prop
        else:
            raise KeyError(f"property '{name}' not found")

    def __setitem__(self, name, value):
        self._set_prop(name, value)

    def __getattr__(self, name):
        prop = self._get_prop(name)
        if prop is not None:
            return prop
        else:
            raise AttributeError(f"property '{name}' not found")

    def __setattr__(self, name, value):
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self._set_prop(name, value)


def _split_text(
    text: str, prop_interpreter: Callable[[str], str], expr_interpreter: Callable[[str], Any]
) -> Iterable[str]:
    """Split input in chunks, applying either the property or expression interpreter based
    on markers.

    Args:
        text: input to process
        prop_interpreter: interpreter for chunks enclosed in curly braces (the braces will be
            included in the input provided)
        expr_interpreter: interpreter for the chunks enclosed in ``%`` markers (the marker will
            *not* be included in the input provided)

    Returns:
        generator of processed text chunks
    """

    in_prop = False
    in_expr = False
    cur_token = ""

    while True:
        if len(text) == 0:
            if in_expr:
                raise ExpressionSubstitutionError(
                    "unterminated expression substitution pattern"
                )
            if in_prop:
                raise PropertySubstitutionError("unterminated property substitution pattern")
            break

        assert not (in_prop and in_expr)

        s, text = text[0], text[1:]
        peek = text[0] if len(text) > 0 else ""

        if s == "{":
            if in_prop:
                raise PropertySubstitutionError(
                    "'{' not allowed inside property substitution patterns"
                )
            elif in_expr:
                cur_token += s
            elif peek == "{":
                cur_token += s
                text = text[1:]
            else:
                yield cur_token
                cur_token = s
                in_prop = True
        elif s == "}":
            if in_prop:
                try:
                    yield str(prop_interpreter(cur_token + s))
                except Exception as exc:
                    raise PropertySubstitutionError() from exc
                cur_token = ""
                in_prop = False
            elif in_expr:
                cur_token += s
            elif peek == "}":
                cur_token += s
                text = text[1:]
            else:
                raise PropertySubstitutionError("closing '}' without matching '{'")
        elif s == "%":
            if in_prop:
                cur_token += s
            elif peek == "%":
                cur_token += s
                text = text[1:]
            elif in_expr:
                try:
                    yield str(expr_interpreter(cur_token))
                except Exception as exc:
                    raise ExpressionSubstitutionError() from exc
                cur_token = ""
                in_expr = False
            else:
                yield cur_token
                cur_token = ""
                in_expr = True
        else:
            cur_token += s

    yield cur_token


def _substitute_expressions(
    text: str, prop_interpreter: Callable[[str], str], expr_interpreter: Callable[[str], Any]
) -> str:
    """Apply substitution with custom interpreter functions for property substitution and for
    expression substitution.

    Used for testing :func:`_split_text`.
    """
    return "".join(_split_text(text, prop_interpreter, expr_interpreter))


def _glob(files: str) -> list[pathlib.Path]:
    return [
        pathlib.Path(file) for file in glob.glob(os.path.expandvars(os.path.expanduser(files)))
    ]


_OS_PATH_SYMBOLS = (
    "abspath",
    "basename",
    "dirname",
    "exists",
    "expanduser",
    "isfile",
    "isdir",
    "splitext",
)


class SubstitutionHelper:
    """Helper class to perform :ref:`property <fundamentals_property_substitution>` and
    :ref:`expression <fundamentals_expression_substitution>` substitution.

    Args:
        state: :class:`State` instance
    """

    def __init__(self, state: State):
        self._property_proxy = _PropertyProxy(state, True, True)
        symtable = {
            **vp.UNITS,
            **{f: getattr(os.path, f) for f in _OS_PATH_SYMBOLS},
            "input": input,
            "glob": _glob,
            "convert_length": vp.convert_length,
            "convert_angle": vp.convert_angle,
            "convert_page_size": vp.convert_page_size,
            "stdin": sys.stdin,
            "Color": vp.Color,
            "prop": self._property_proxy,
            "lprop": _PropertyProxy(state, False, True),
            "gprop": _PropertyProxy(state, True, False),
        }
        # disabling numpy as its math functions such as `ceil` do not convert to int.
        self._interpreter = asteval.Interpreter(
            usersyms=symtable,
            readonly_symbols=symtable.keys() - vp.UNITS.keys(),
            use_numpy=False,
        )

    @property
    def symtable(self) -> dict[str, Any]:
        return self._interpreter.symtable

    def substitute(self, text: str) -> str:
        """Apply :ref:`property <fundamentals_property_substitution>` and
        :ref:`expression <fundamentals_expression_substitution>` substitution.

        Args:
            text: user input

        Returns:
            fully substituted text
        """
        return _substitute_expressions(
            text,
            lambda s: s.format_map(self._property_proxy),
            lambda s: self._interpreter(s, raise_errors=True, show_errors=False),
        )



from __future__ import annotations

import random

import numpy as np
import pnoise

from .geometry import interpolate
from .model import LineCollection

__all__ = ["squiggles"]


def squiggles(
    lines: LineCollection, ampl: float, period: float, quantization: float
) -> LineCollection:
    """Apply a squiggle filter to a :class:`LineCollection`.

    This filter first densely resample the input lines (based on the ``quantization``
    parameter), and then applies a 2D-Perlin-noise-based displacement to all points.

    For small values of amplitude (~2px) and period (~12px), this filter gives a "shaky-hand"
    style to the lines. Larger values of amplitude (~60px) and period (~400px) result in a
    a smoother, liquid-like effect.

    Args:
        lines: input lines
        ampl: squiggle amplitude (px)
        period: squiggle period (px)
        quantization: quantization (px)

    Returns:
        filtered lines
    """

    # link noise seed to global seed
    noise = pnoise.Noise()
    noise.seed(random.randint(0, 2**16))

    freq = 1.0 / period
    new_lines = lines.clone()
    for line in lines:
        line = interpolate(line, step=quantization)
        perlin_x = noise.perlin(
            freq * line.real, freq * line.imag, np.zeros_like(line.real), grid_mode=False
        )
        perlin_y = noise.perlin(
            freq * line.real, freq * line.imag, 1000 * np.ones_like(line.real), grid_mode=False
        )
        new_lines.append(line + ampl * 2.0 * ((perlin_x - 0.5) + (perlin_y - 0.5) * 1j))

    return new_lines



from __future__ import annotations

import moderngl
from PIL import Image

import vpype as vp

from .._scales import UnitType
from ..engine import DEFAULT_PEN_OPACITY, DEFAULT_PEN_WIDTH, Engine, ViewMode


class ImageRenderer:
    """Viewer engine wrapper class to render to a :class:`PIL.Image.Image` instance.

    Example:

        >>> doc = vp.Document()
        # populate doc
        >>> renderer = ImageRenderer((640, 480))
        >>> renderer.engine.document = doc
        >>> renderer.engine.fit_to_viewport()
        >>> img = renderer.render()
    """

    def __init__(self, size: tuple[int, int]):
        """Constructor.

        Args:
            size: image size
        """
        ctx = moderngl.create_standalone_context()
        self._fbo = ctx.simple_framebuffer(size)
        self.engine = Engine()
        self.engine.post_init(ctx, *size)

    def render(self) -> Image:
        """Render to a :class:`PIL.Image.Image` instance.

        Returns:
            :class:`PIL.Image.Image`: the rendered mage
        """
        self._fbo.use()
        self.engine.render()
        return Image.frombytes("RGB", self._fbo.size, self._fbo.read(), "raw", "RGB", 0, -1)


def render_image(
    document: vp.Document,
    size: tuple[int, int] = (512, 512),
    view_mode: ViewMode = ViewMode.PREVIEW,
    show_pen_up: bool = False,
    show_points: bool = False,
    default_pen_width: float = DEFAULT_PEN_WIDTH,
    default_pen_opacity: float = DEFAULT_PEN_OPACITY,
    override_pen_width: bool = False,
    override_pen_opacity: bool = False,
    show_ruler: bool = False,
    pixel_factor: float = 1.0,
    unit_type: UnitType = UnitType.METRIC,
    scale: float | None = None,
    origin: tuple[float, float] | None = None,
) -> Image:
    """Render a :class:`vpype.Document` instance as a Pillow :class:`PIL.Image.Image`.

    By default, the document is scaled and offset to entirely fit in the image. This behaviour
    can be overridden using the ``scale`` and ``offset`` arguments.

    Args:
        document: document to render
        size: output image size
        view_mode: :class:`ViewMode` to use
        show_pen_up: render pen-up trajectories
        show_points: render points
        default_pen_width: pen width (``ViewMode.PREVIEW`` only)
        default_pen_opacity: pen opacity (``ViewMode.PREVIEW`` only)
        override_pen_width: override pen width property (``ViewMode.PREVIEW`` only)
        override_pen_opacity: override pen opacity property (``ViewMode.PREVIEW`` only)
        show_ruler: display the rulers
        pixel_factor: pixel factor (HiDPI screen support)
        unit_type: type of unit to use for the ruler
        scale: manually set scale
        origin: manually set origin

    Returns:
        :class:`PIL.Image.Image`: the rendered image
    """

    renderer = ImageRenderer(size)

    renderer.engine.document = document
    renderer.engine.view_mode = view_mode
    renderer.engine.show_pen_up = show_pen_up
    renderer.engine.show_points = show_points
    renderer.engine.default_pen_width = default_pen_width
    renderer.engine.default_pen_opacity = default_pen_opacity
    renderer.engine.override_pen_width = override_pen_width
    renderer.engine.override_pen_opacity = override_pen_opacity
    renderer.engine.show_rulers = show_ruler
    renderer.engine.pixel_factor = pixel_factor
    renderer.engine.unit_type = unit_type

    renderer.engine.fit_to_viewport()
    if scale is not None:
        renderer.engine.scale = scale
    if origin is not None:
        renderer.engine.origin = origin

    return renderer.render()



from __future__ import annotations

import copy
import importlib.metadata
import logging
import os
import random
import shlex
import sys
import traceback
from typing import TYPE_CHECKING, Any, Callable, Iterable, TextIO, Union, cast

import click
import numpy as np

import vpype as vp

from .decorators import global_processor
from .state import State

__all__ = ("cli", "execute", "begin", "end", "execute_processors", "ProcessorType")


ProcessorType = Union[
    Callable,
    "BeginBlock",
    "EndBlock",
]


class GroupedGroup(click.Group):
    """Custom group class which implements command grouping in --help display.

    Based on Stephen Rauch's excellent answer: https://stackoverflow.com/a/58770064/229511
    """

    def command(self, *args, **kwargs):
        """Gather the command help groups"""
        help_group = kwargs.pop("group", None)
        decorator = super().command(*args, **kwargs)

        def wrapper(f):
            cmd = decorator(f)
            cmd.help_group = help_group
            return cmd

        return wrapper

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            longest = max(len(cmd[0]) for cmd in commands)
            limit = formatter.width - 6 - longest

            groups = {}
            for subcommand, cmd in commands:
                help_text = cmd.get_short_help_str(limit)
                subcommand += " " * (longest - len(subcommand))
                groups.setdefault(getattr(cmd, "help_group", "Unknown"), []).append(
                    (subcommand, help_text)
                )

            with formatter.section("Commands"):
                for group_name, rows in groups.items():
                    with formatter.section(group_name):
                        formatter.write_dl(rows)

    def main(self, args=None, **extra):
        """Let's get a chance to pre-process the argument list for include options."""
        if args is None:
            args = sys.argv[1:]
        return super().main(args=preprocess_argument_list(args), **extra)


class _BrokenCommand(click.Command):  # pragma: no cover
    """Rather than completely crash the CLI when a broken plugin is loaded, this
    class provides a modified help message informing the user that the plugin is
    broken and they should contact the owner.  If the user executes the plugin
    or specifies `--help` a traceback is reported showing the exception the
    plugin loader encountered.
    """

    def __init__(self, name):
        """Define the special help messages after instantiating a `click.Command()`."""

        click.Command.__init__(self, name)
        util_name = os.path.basename(sys.argv and sys.argv[0] or __file__)
        self.help = (
            "\nWarning: entry point could not be loaded. Contact "
            "its author for help.\n\n\b\n" + traceback.format_exc()
        )
        self.short_help = "\u2020 Warning: could not load plugin. See `{} {} --help`.".format(
            util_name,
            self.name,
        )

    def invoke(self, ctx):
        """Print the traceback instead of doing nothing."""
        click.echo(self.help, color=ctx.color)
        ctx.exit(1)

    def parse_args(self, ctx, args):
        return args


_PLUGINS_LOADED = False


# noinspection PyUnusedLocal
@click.group(cls=GroupedGroup, chain=True, invoke_without_command=True)
@click.version_option(version=vp.__version__, message="%(prog)s %(version)s")
@click.option("-h", "--help", "help_flag", is_flag=True, help="Show this message and exit.")
@click.option("-v", "--verbose", count=True)
@click.option("-I", "--include", type=click.Path(), help="Load commands from a command file.")
@click.option(
    "-H",
    "--history",
    is_flag=True,
    help="Record this command in a `vpype_history.txt` file in the current directory.",
)
@click.option("-s", "--seed", type=int, help="Specify the RNG seed.")
@click.option(
    "-c", "--config", type=click.Path(exists=True), help="Load an additional config file."
)
@click.pass_context
def cli(
    ctx: click.Context,
    help_flag: bool,
    verbose: int,
    include: bool,
    history: bool,
    seed: int,
    config: str,
):
    """Execute the sequence of commands passed in argument.

    The available commands are listed below. Information on each command may be obtained using:

        vpype COMMAND --help

    Some of vpype's commands or plug-ins may rely on a random number generator (RNG). By
    default, vpype's RNG is seeded with the current time, such as to produce
    pseudo-random behaviour. The seed can instead be set to a specific value (using
    the `--seed` option) when reproducible behaviour is needed. For example, the following
    always yields the exact same result:

        vpype -s 0 random show

    Include files (commonly named with the `.vpy` extension) can be used instead passing
    commands in the command line, e.g.:

        vpype read input.svg -I my_post_processing.vpy write.output.svg

    Some commands and plug-in can be customized via a TOML configuration file. If a file named
    `.vpype.toml` exists at the root of the user directory, vpype will automatically load it.
    Alternatively, a custom configuration file may be loaded with the `--config` option, e.g.:

        vpype -c my_plotter_config.toml read input.svg write -d my_plotter output.hpgl

    When using the `--history` option, vpype will append its invocation (i.e. the full command
    line) in a `vpype_history.txt` file in the current directory (creating it if necessary).
    This may be useful to easily keep a trace of how project might have been created or
    post-processed with vpype.

    By default, vpype verbosity is low. It may be increased by using the `-v` option once or
    twice to increase verbosity to info, respectively debug level, e.g.:

        vpype -vv [...]

    Refer to the documentation at https://vpype.readthedocs.io/ for more information.
    """

    logging.basicConfig()
    if verbose == 0:
        logging.getLogger().setLevel(logging.WARNING)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose > 1:
        logging.getLogger().setLevel(logging.DEBUG)

    # Plug-in loading logic. This approach is preferred because:
    # 1) Deferred plug-in loading avoid circular import between vpype and vpype_cli when plug-
    #    in uses deprecated APIs.
    # 2) Avoids the PyCharm type error with CliRunner.invoke()
    global _PLUGINS_LOADED
    if not _PLUGINS_LOADED:
        _PLUGINS_LOADED = True
        for entry_point in importlib.metadata.entry_points().get("vpype.plugins", []):
            # noinspection PyBroadException
            try:
                cast(click.Group, ctx.command).add_command(entry_point.load())
            except Exception:
                # Catch this so a busted plugin doesn't take down the CLI.
                # Handled by registering a dummy command that does nothing
                # other than explain the error.
                cast(click.Group, ctx.command).add_command(_BrokenCommand(entry_point.name))

    # Manual handling of the help to work around circular import issues.
    # Background: when importing plug-ins in the style of `click-plugin` (decorator, so plug-
    # ins are loaded during the loading of `cli` itself), plug-in may not import things from
    # `vpype_cli` since it is still partially loaded. Plug-ins are thus loaded when `cli` is
    # actually executed (see previous lines). As a result, the Click's default behaviour for
    # handling `--help` (i.e. print and exit *before* even executing `cli`) is unable to list
    # the plug-ins. This is addressed by manually handling the top-level `--help` parameter
    # *after* plug-ins are loaded.
    if help_flag:
        print(ctx.command.get_help(ctx))
        sys.exit(0)

    # We use the command string as context object, mainly for the purpose of the `write`
    # command. This is a bit of a hack, and will need to be updated if we ever need more state
    # to be passed around (probably State should go in there!)
    cmd_string = "vpype " + " ".join(shlex.quote(arg) for arg in sys.argv[1:]) + "\n"
    ctx.obj = cmd_string

    if history:
        with open("vpype_history.txt", "a") as fp:
            fp.write(cmd_string)

    if seed is None:
        seed = np.random.randint(2**31)
        logging.info(f"vpype: no seed provided, using {seed}")
    np.random.seed(seed)
    random.seed(seed)

    if config is not None:
        vp.config_manager.load_config_file(config)


# this is somehow needed to make PyCharm happy with runner.invoke(cli, ...)
if TYPE_CHECKING:  # pragma: no cover
    cli = cast(GroupedGroup, cli)


# noinspection PyUnusedLocal
@cli.result_callback()
def process_pipeline(processors, help_flag, verbose, include, history, seed, config):
    execute_processors(processors, State())


def execute_processors(processors: Iterable[ProcessorType], state: State) -> None:
    """Execute a sequence of processors to generate a Document structure. For block handling,
    we use a recursive approach. Only top-level blocks are extracted and processed by block
    processors, which, in turn, recursively call this function.

    Args:
        processors: iterable of processors
        state: state structure

    Returns:
        generated geometries
    """

    outer_processors: list[Any] = []  # gather commands outside of top-level blocks
    top_level_processors: list[Any] = []  # gather commands inside of top-level blocks
    block = None  # save the current top-level block's block layer_processor
    nested_count = 0  # block depth counter
    expect_block = False  # set to True by `begin` command

    for proc in processors:
        if getattr(proc, "__vpype_block_processor__", False):
            if not expect_block:
                # `begin` was omitted
                nested_count += 1
            else:
                expect_block = False

            # if we in a top level block, we save the block layer_processor
            # (nested block are ignored for the time being)
            if nested_count == 1:
                block = proc
            else:
                top_level_processors.append(proc)
        elif expect_block:
            raise click.BadParameter("A block command must always follow 'begin'")
        elif isinstance(proc, BeginBlock):
            # entering a block
            nested_count += 1
            expect_block = True
        elif isinstance(proc, EndBlock):
            if nested_count < 1:
                raise click.BadParameter(
                    "A 'end' command has no corresponding 'begin' command"
                )

            nested_count -= 1

            if nested_count == 0:
                # We're closing a top level block. The top-level sequence [BeginBlock,
                # block_processor, *top_level_processors, EndBlock] is now replaced by a
                # placeholder closure that will execute the corresponding block processor on
                # the top_level_processors sequence.
                #
                # Note: we use the default argument trick to copy the *current* value of
                # block and top_level_processor "inside" the placeholder function.

                # noinspection PyShadowingNames
                def block_processor_placeholder(
                    state: State, block=block, processors=tuple(top_level_processors)
                ) -> State:
                    return cast(Callable, block)(state, processors)

                outer_processors.append(block_processor_placeholder)

                # reset the top level layer_processor list
                top_level_processors = list()
            else:
                top_level_processors.append(proc)
        else:
            # this is a 'normal' layer_processor, we can just add it to the top of the stack
            if nested_count == 0:
                outer_processors.append(proc)
            else:
                top_level_processors.append(proc)

    # at this stage, the stack must have a single frame, otherwise we're missing end commands
    if nested_count > 0:
        raise click.ClickException("An 'end' command is missing")

    # the (only) frame's processors should now be flat and can be chain-called
    for proc in outer_processors:
        cast(Callable, proc)(state)


class BeginBlock:
    pass


@cli.command(group="Block control")
def begin():
    """Marks the start of a block.

    A `begin` command must be followed by a block processor command (eg. `grid` or `repeat`),
    which indicates how the block is processed. Blocks must be ended by a `end` command.

    Blocks can be nested.
    """
    return BeginBlock()


class EndBlock:
    pass


@cli.command(group="Block control")
def end():
    """Marks the end of a block."""
    return EndBlock()


def extract_arguments(f: TextIO) -> list[str]:
    """Read the content of a file-like object and extract the corresponding argument list.

    Everything following a '#' is ignored until end of line. Any whitespace is considered
    to separate arguments. Single and double quote are honored, i.e. content becomes an
    argument (but quote are removed).

    Args:
        f: file-like object

    Returns:
        list of argument extracted from input
    """
    args = []
    for line in f.readlines():
        args.extend(shlex.split(line, comments=True))
    return args


def preprocess_argument_list(args: list[str], cwd: str | None = None) -> list[str]:
    """Preprocess an argument list, replacing 'include' options by the corresponding file's
    content.

    Include options are either '-I' or '--include', and must be followed by a file path. This
    behaviour is recursive, e.g. a file could contain an include statement as well.

    Args:
        args: argument list
        cwd:  current working directory, used as reference for relative file paths (use
            actual current working directory if None)

    Returns:
        preprocessed list or argument
    """

    if cwd is None:
        cwd = os.getcwd()

    result = []

    while len(args) > 0:
        arg = args.pop(0)

        if arg == "-I" or arg == "--include":
            if len(args) == 0:
                raise click.ClickException("include option must be followed by a file path")
            else:
                # include statement in files are relative to that file, so we need to
                # provide the file's path to ourselves
                file_path = args.pop(0)
                if not os.path.isabs(file_path):
                    file_path = os.path.join(cwd, file_path)
                dir_path = os.path.dirname(file_path)

                with open(file_path) as f:
                    result.extend(preprocess_argument_list(extract_arguments(f), dir_path))
        else:
            result.append(arg)

    return result


def execute(
    pipeline: str, document: vp.Document | None = None, global_opt: str = ""
) -> vp.Document:
    """Execute a vpype pipeline.

    This function serves as a Python API to vpype's pipeline. It can be used from a regular
    Python script (as opposed to the ``vpype`` CLI which must be used from a console or via
    :func:`os.system`.

    If a :class:`vpype.Document` instance is provided, it will be preloaded in the pipeline
    before the first command executes. The pipeline's content after the last command is
    returned as a :class:`vpype.Document` instance.

    Examples:

        Read a SVG file, optimize it and return the result as a :class:`vpype.Document`
        instance::

            >>> doc = execute("read input.svg linemerge linesimplify linesort")

        Optimize and save a :class:`vpype.Document` instance::

            >>> doc = vp.Document()
            >>> # populate `doc` with some graphics
            >>> execute("linemerge linesimplify linesort write output.svg", doc)

    Args:
        pipeline: vpype pipeline as would be used with ``vpype`` CLI
        document: if provided, is perloaded in the pipeline before the first command executes
        global_opt: global CLI option (e.g. "--verbose")

    Returns:
        pipeline's content after the last command executes
    """

    if document:
        # noinspection PyUnusedLocal
        @cli.command()
        @global_processor
        def vsketchinput(doc):
            return copy.deepcopy(document)

    out_doc = vp.Document()

    @cli.command()
    @global_processor
    def vsketchoutput(doc):
        nonlocal out_doc
        out_doc = doc
        return doc

    args = " ".join(
        [global_opt, ("vsketchinput " if document else ""), pipeline, "vsketchoutput"]
    )
    cli.main(prog_name="vpype", args=shlex.split(args), standalone_mode=False)
    return out_doc



"""This module contains vpype core and its API."""

from .config import *
from .filters import *
from .geometry import *
from .io import *
from .line_index import *
from .metadata import *
from .model import *
from .primitives import *
from .text import *
from .utils import *


def _get_version() -> str:
    from importlib.metadata import version

    return version(__name__)


__version__ = _get_version()



from __future__ import annotations

import os
import signal
import socket
from contextlib import contextmanager
from typing import Callable

from PySide6 import QtNetwork
from PySide6.QtGui import QAction, QActionGroup, QGuiApplication, QIcon, QPalette


def load_icon(path: str) -> QIcon:
    path = os.path.abspath(
        os.path.dirname(__file__) + os.path.sep + "resources" + os.path.sep + path
    )

    # check if dark mode is enabled
    app = QGuiApplication.instance()

    if app is None or not isinstance(app, QGuiApplication):
        raise EnvironmentError("no Qt application available")

    base_color = app.palette().color(QPalette.Base)
    if base_color.lightnessF() < 0.5:
        file, ext = os.path.splitext(path)
        path = file + "-dark" + ext

    icon = QIcon(path)
    return icon


class PenWidthActionGroup(QActionGroup):
    widths = [
        0.05,
        0.1,
        0.15,
        0.2,
        0.25,
        0.3,
        0.35,
        0.4,
        0.45,
        0.5,
        0.6,
        0.7,
        0.8,
        0.9,
        1.0,
        1.2,
    ]

    def __init__(self, current: float = 0.3, parent=None):
        super().__init__(parent)

        if current not in self.widths:
            self.widths.append(current)
        for w in sorted(self.widths):
            act = self.addAction(QAction(f"{w} mm"))
            act.setCheckable(True)
            act.setChecked(w == current)
            act.setData(w)


class PenOpacityActionGroup(QActionGroup):
    opacities = [
        0.3,
        0.5,
        0.7,
        0.8,
        0.9,
        0.95,
        1,
    ]

    def __init__(self, current: float = 0.8, parent=None):
        super().__init__(parent)

        if current not in self.opacities:
            self.opacities.append(current)
        for w in sorted(self.opacities):
            act = self.addAction(QAction(f"{int(w*100)}%"))
            act.setCheckable(True)
            act.setChecked(w == current)
            act.setData(w)


class SignalWatchdog(QtNetwork.QAbstractSocket):
    """This object notify PySide6's event loop of an incoming signal and makes it process it.

    The python interpreter flags incoming signals and triggers the handler only upon the next
    bytecode is processed. Since PySide6's C++ event loop function never/rarely returns when
    the UX is in the background, the Python interpreter doesn't have a chance to run and call
    the handler.

    From: https://stackoverflow.com/a/65802260/229511 and
    https://stackoverflow.com/a/37229299/229511
    """

    def __init__(self):
        # noinspection PyTypeChecker
        super().__init__(QtNetwork.QAbstractSocket.SctpSocket, None)  # type: ignore
        self.writer, self.reader = socket.socketpair()
        self.writer.setblocking(False)
        signal.set_wakeup_fd(self.writer.fileno())
        self.setSocketDescriptor(self.reader.fileno())
        self.readyRead.connect(lambda: None)  # type: ignore


@contextmanager
def set_sigint_handler(handler: Callable):
    original_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, handler)
    # noinspection PyUnusedLocal
    watchdog = SignalWatchdog()
    try:
        yield
    finally:
        signal.signal(signal.SIGINT, original_handler)
        del watchdog



"""This module implements vpype's CLI interface and the :func:`execute() <vpype_cli.execute>`
function.
"""
# register all commands
from .blocks import *
from .cli import *
from .debug import *
from .decorators import *
from .eval import *
from .filters import *
from .frames import *
from .generators import *
from .layerops import *
from .metadata import *
from .operations import *
from .primitives import *
from .read import *
from .script import *
from .show import *
from .state import *
from .text import *
from .transforms import *
from .types import *
from .write import *



from __future__ import annotations

import click

import vpype as vp

from .cli import cli
from .decorators import layer_processor
from .types import LengthType


@cli.command(group="Filters")
@click.option(
    "-a",
    "--amplitude",
    type=LengthType(),
    default="0.5mm",
    help="Amplitude of the noise-based displacement (default: 0.5mm).",
)
@click.option(
    "-p",
    "--period",
    type=LengthType(),
    default="3mm",
    help="Period of the noise-based displacement (default: 3mm).",
)
@click.option(
    "-q",
    "--quantization",
    type=LengthType(),
    default="0.05mm",
    help="Maximum segment size used for the resampling (default: 0.05mm).",
)
@layer_processor
def squiggles(lines: vp.LineCollection, amplitude: float, period: float, quantization: float):
    """Apply a squiggle filter to the geometries.

    This filter works by first resampling the input lines, and then applying a random
    displacement to all points. This displacement is based on a 2D Perlin noise field with
    adjustable amplitude and period.

    The default values of amplitude and period give a "shaky-hand" style to the geometries.
    Larger values of amplitude (~15mm) and period (~10cm) result in a a smoother, liquid-like
    effect.
    """

    return vp.squiggles(lines, amplitude, period, quantization)



from __future__ import annotations

import os
import pathlib
from typing import Any, Dict, Tuple, cast

import cachetools
import cachetools.keys
import moderngl as mgl
import numpy as np
from PIL import Image

ColorType = Tuple[float, float, float, float]


@cachetools.cached(
    cache=cast(Dict[Any, Any], {}),
    key=lambda name, ctx: cachetools.keys.hashkey((name, id(ctx))),
)
def load_program(name: str, ctx: mgl.Context) -> mgl.Program:
    """Create a program based on a shader name prefix.

    This function automatically load all shader type based on file existence, looking for
    _vertex, _fragment and _geometry suffices in the ``shaders`` directory.

    The same program instance is returned for identical combination of ``name``/``_ctx``.

    Args:
         name: shader file name prefix
         ctx: context

    Returns:
        the loaded program
    """

    def _load_shader(path: str) -> str | None:
        try:
            with open(path) as fp:
                return fp.read()
        except OSError:
            return None

    full_path = os.path.dirname(__file__) + os.path.sep + "shaders" + os.path.sep + name

    return ctx.program(
        vertex_shader=_load_shader(full_path + "_vertex.glsl"),
        fragment_shader=_load_shader(full_path + "_fragment.glsl"),
        geometry_shader=_load_shader(full_path + "_geometry.glsl"),
    )


@cachetools.cached(
    cache=cast(Dict[Any, Any], {}),
    key=lambda name, ctx, size, components: cachetools.keys.hashkey(
        (name, id(ctx), size, components)
    ),
)
def load_texture_array(
    name: str, ctx: mgl.Context, size: tuple[int, int, int], components: int = 4
) -> mgl.TextureArray:
    texture_path = pathlib.Path(__file__).parent / "resources" / name
    img = Image.open(str(texture_path))
    texture = ctx.texture_array(size, components, data=img.convert("RGBA").tobytes())
    texture.build_mipmaps()
    return texture


def orthogonal_projection_matrix(
    left: float, right: float, bottom: float, top: float, near: float, far: float, dtype=None
) -> np.ndarray:
    """Creates an orthogonal projection matrix."""

    rml = right - left
    tmb = top - bottom
    fmn = far - near

    a = 2.0 / rml
    b = 2.0 / tmb
    c = -2.0 / fmn
    tx = -(right + left) / rml
    ty = -(top + bottom) / tmb
    tz = -(far + near) / fmn

    # GLSL is column major, thus the transpose
    return np.array(
        (
            (a, 0.0, 0.0, 0.0),
            (0.0, b, 0.0, 0.0),
            (0.0, 0.0, c, 0.0),
            (tx, ty, tz, 1.0),
        ),
        dtype=dtype,
    )



from __future__ import annotations

import click

import vpype as vp

from .cli import cli
from .decorators import generator
from .types import LengthType, TextType


@cli.command(group="Text")
@click.argument("string", type=TextType())
@click.option(
    "-f", "--font", type=click.Choice(vp.FONT_NAMES), default="futural", help="Font to use."
)
@click.option("-s", "--size", type=LengthType(), default=18, help="Text size (default: 18).")
@click.option("-w", "--wrap", type=LengthType(), help="Wrap to provided width.")
@click.option("-j", "--justify", is_flag=True, help="Justify text block (wrap-mode only).")
@click.option(
    "-p",
    "--position",
    nargs=2,
    type=LengthType(),
    default=[0, 0],
    help="Position of the text (default: 0, 0).",
)
@click.option(
    "-a",
    "--align",
    type=click.Choice(["left", "right", "center"]),
    default="left",
    help="Text alignment with respect to position (default: left).",
)
@generator
def text(
    string: str,
    font: str,
    size: float,
    wrap: float | None,
    justify: float,
    position: tuple[float, float],
    align: str,
):
    """Generate text using Hershey fonts.

    By default, `text` generates a single line of text. Automatic text wrapping can be enabled
    using the the `--wrap` option.

    In normal mode, the text starts at coordinate (0, 0) and expends right. This can be
    controlled using the `--align` option. By passing "center", the text is centered around
    (0, 0). With "right", the text expends left of (0, 0).

    In wrap mode, the text start at (0, 0) and expends left until it reach the specified width.
    The `--align` option controls how the text is laid out within the column and behaves as
    typically expected.

    To start the text at the different location than (0, 0), use the `--position` option.
    """

    # skip if text is empty
    if string.strip() == "":
        return vp.LineCollection()

    if wrap:
        lc = vp.text_block(
            string, font_name=font, width=wrap, size=size, align=align, justify=justify
        )
    else:
        lc = vp.text_line(string, font_name=font, size=size, align=align)

    lc.translate(position[0], position[1])

    return lc



from __future__ import annotations

import click

import vpype as vp

from .cli import cli
from .decorators import generator
from .types import AngleType, LengthType

__all__ = ("arc", "circle", "ellipse", "line", "rect")


@cli.command(group="Primitives")
@click.argument("x0", type=LengthType())
@click.argument("y0", type=LengthType())
@click.argument("x1", type=LengthType())
@click.argument("y1", type=LengthType())
@generator
def line(x0: float, y0: float, x1: float, y1: float) -> vp.LineCollection:
    """Generate a single line.

    The line starts at (X0, Y0) and ends at (X1, Y1). All arguments understand supported units.
    """
    return vp.LineCollection([vp.line(x0, y0, x1, y1)])


@cli.command(group="Primitives")
@click.argument("x", type=LengthType())
@click.argument("y", type=LengthType())
@click.argument("width", type=LengthType())
@click.argument("height", type=LengthType())
@click.option(
    "--radii",
    "-r",
    type=LengthType(),
    nargs=4,
    default=(0, 0, 0, 0),
    help="Top-left, top-right, bottom-right and bottom-left corner radii.",
)
@click.option(
    "-q",
    "--quantization",
    type=LengthType(),
    default="1mm",
    help="Maximum length of segments approximating the rounded angles.",
)
@generator
def rect(
    x: float,
    y: float,
    width: float,
    height: float,
    radii: tuple[float, float, float, float],
    quantization: float,
) -> vp.LineCollection:
    """Generate a rectangle, with optional rounded angles.

    The rectangle is defined by its top left corner (X, Y) and its width and height.

    Examples:

        Straight-angle rectangle:

            vpype rect 10cm 10cm 3cm 2cm show

        Rounded-angle rectangle:

            vpype rect --radii 5mm 5mm 5mm 5mm 10cm 10cm 3cm 2cm show

        Rounded-angle rectangle with quantization control:

            vpype rect --quantization 0.1mm --radii 5mm 5mm 5mm 5mm 10cm 10cm 3cm 2cm show
    """
    return vp.LineCollection([vp.rect(x, y, width, height, *radii, quantization)])


@cli.command(group="Primitives")
@click.argument("x", type=LengthType())
@click.argument("y", type=LengthType())
@click.argument("rw", type=LengthType())
@click.argument("rh", type=LengthType())
@click.argument("start", type=AngleType())
@click.argument("stop", type=AngleType())
@click.option(
    "-q",
    "--quantization",
    type=LengthType(),
    default="1mm",
    help="Maximum length of segments approximating the arc.",
)
@generator
def arc(
    x: float, y: float, rw: float, rh: float, start: float, stop: float, quantization: float
):
    """Generate lines approximating a circular arc.

    The arc is centered on (X, Y) and has a radius of R and spans counter-clockwise from START
    to STOP angles. Angular values of zero refer to east of unit circle and positive values
    extend counter-clockwise.

    Angles are in degree by default, but alternative CSS units such as "rad" or "grad" may be
    provided.
    """
    return vp.LineCollection([vp.arc(x, y, rw, rh, start, stop, quantization)])


@cli.command(group="Primitives")
@click.argument("x", type=LengthType())
@click.argument("y", type=LengthType())
@click.argument("r", type=LengthType())
@click.option(
    "-q",
    "--quantization",
    type=LengthType(),
    default="1mm",
    help="Maximum length of segments approximating the circle.",
)
@generator
def circle(x: float, y: float, r: float, quantization: float):
    """Generate lines approximating a circle.

    The circle is centered on (X, Y) and has a radius of R.
    """

    return vp.LineCollection([vp.circle(x, y, r, quantization)])


@cli.command(group="Primitives")
@click.argument("x", type=LengthType())
@click.argument("y", type=LengthType())
@click.argument("w", type=LengthType())
@click.argument("h", type=LengthType())
@click.option(
    "-q",
    "--quantization",
    type=LengthType(),
    default="1mm",
    help="Maximum length of segments approximating the ellipse.",
)
@generator
def ellipse(x: float, y: float, w: float, h: float, quantization: float):
    """Generate lines approximating an ellipse.

    The ellipse is centered on (X, Y), with a half-width of W and a half-height of H.
    """

    return vp.LineCollection([vp.ellipse(x, y, w, h, quantization)])



from __future__ import annotations

import logging
from typing import Any, ClassVar

import click

import vpype as vp

from .state import State, _DeferredEvaluator


class _DeferredEvaluatorType(click.ParamType):
    """Base class for types which rely on a deferred evaluator.

    Sub-classes must set the value of ``_evaluator_class``.
    """

    _evaluator_class: ClassVar = _DeferredEvaluator

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            return self.__class__._evaluator_class(
                value, param.human_readable_name, *self._args, **self._kwargs
            )
        else:
            return super().convert(value, param, ctx)


class TextType(_DeferredEvaluatorType):
    """:class:`click.ParamType` sub-class to automatically perform
    :ref:`property substitution <fundamentals_property_substitution>` on user input.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("text", type=vpype_cli.TextType())
        ... @vpype_cli.generator
        ... def my_command(text: str):
        ...     pass
    """

    class _TextDeferredEvaluator(_DeferredEvaluator):
        def evaluate(self, state: State) -> str:
            return state.substitute(self._text)

    name = "text"
    _evaluator_class = _TextDeferredEvaluator


class IntegerType(_DeferredEvaluatorType):
    """:class:`click.ParamType` sub-class to automatically perform
    :ref:`property substitution <fundamentals_property_substitution>` on user input.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("number", type=vpype_cli.IntegerType())
        ... @vpype_cli.generator
        ... def my_command(number: int):
        ...     pass
    """

    class _IntegerDeferredEvaluator(_DeferredEvaluator):
        def evaluate(self, state: State) -> int:
            return int(state.substitute(self._text))

    name = "number"
    _evaluator_class = _IntegerDeferredEvaluator


class FloatType(_DeferredEvaluatorType):
    """:class:`click.ParamType` sub-class to automatically perform
    :ref:`property substitution <fundamentals_property_substitution>` on user input.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("number", type=vpype_cli.FloatType())
        ... @vpype_cli.generator
        ... def my_command(number: float):
        ...     pass
    """

    class _FloatDeferredEvaluator(_DeferredEvaluator):
        def evaluate(self, state: State) -> float:
            return float(state.substitute(self._text))

    name = "number"
    _evaluator_class = _FloatDeferredEvaluator


class LengthType(_DeferredEvaluatorType):
    """:class:`click.ParamType` sub-class to automatically converts a user-provided lengths
    into CSS pixel units.


    User-provided length strings may contains units which are converted using
    :func:`convert_length`. :ref:`Property substitution <fundamentals_property_substitution>`
    is perfomred as well.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("x", type=vpype_cli.LengthType())
        ... @click.option("-o", "--option", type=vpype_cli.LengthType(), default="1mm")
        ... @vpype_cli.generator
        ... def my_command(x: float, option: float):
        ...     pass
    """

    class _LengthDeferredEvaluator(_DeferredEvaluator):
        def evaluate(self, state: State) -> float:
            return vp.convert_length(state.substitute(self._text))

    name = "length"
    _evaluator_class = _LengthDeferredEvaluator


class AngleType(_DeferredEvaluatorType):
    """:class:`click.ParamType` sub-class to automatically converts a user-provided angle.

    User-provided angle strings may contain units and are converted into a value in degrees.
    This class uses :func:`convert_angle` internally.
    :ref:`Property substitution <fundamentals_property_substitution>` is perfomred as well.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("angle", type=vpype_cli.AngleType())
        ... @vpype_cli.generator
        ... def my_command(angle: float):
        ...     pass
    """

    class _AngleDeferredEvaluator(_DeferredEvaluator):
        def evaluate(self, state: State) -> float:
            return vp.convert_angle(state.substitute(self._text))

    name = "angle"
    _evaluator_class = _AngleDeferredEvaluator


class PageSizeType(_DeferredEvaluatorType):
    """:class:`click.ParamType` sub-class to automatically converts a user-provided page size.

    User-provided page size strings are converted into a tuple of float in CSS pixel units.
    See :func:`convert_page_size` for information on the page size descriptor syntax.
    :ref:`Property substitution <fundamentals_property_substitution>` is perfomred as well.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("fmt", type=vpype_cli.PageSizeType())
        ... @vpype_cli.generator
        ... def my_command(fmt: tuple[float, float]):
        ...     pass
    """

    class _PageSizeDeferredEvaluator(_DeferredEvaluator):
        def evaluate(self, state: State) -> tuple[float, float]:
            return vp.convert_page_size(state.substitute(self._text))

    name = "pagesize"
    _evaluator_class = _PageSizeDeferredEvaluator


class _DelegatedDeferredEvaluatorType(click.ParamType):
    """Base class for types which delegate the deferred evaluation to another type class.

    Sub-class will behave like the delegate class (in particular __init__ will accept the same
    arguments), but the delegate class will only be created upon deferred evaluation of user
    input.
    """

    class _DelegatedDeferredEvaluator(_DeferredEvaluator):
        def __init__(self, text: str, param_name: str, cls: type, *args, **kwargs):
            super().__init__(text, param_name)
            self._cls = cls
            self._args = args
            self._kwargs = kwargs

        def evaluate(self, state: State) -> Any:
            delegated = self._cls(*self._args, **self._kwargs)
            return delegated.convert(state.substitute(self._text), None, None)

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def convert(self, value, param, ctx):
        if isinstance(value, str):
            return self.__class__._DelegatedDeferredEvaluator(
                value,
                param.human_readable_name,
                self.__class__._delegate_class,
                *self._args,
                **self._kwargs,
            )
        else:
            return super().convert(value, param, ctx)

    _delegate_class: ClassVar = click.ParamType


class PathType(_DelegatedDeferredEvaluatorType):
    """:class:`click.Path` clone which performs substitution on input."""

    name = "file"
    _delegate_class = click.Path


class FileType(_DelegatedDeferredEvaluatorType):
    """:class:`click.File` clone which performs substitution on input."""

    name = "file"
    _delegate_class = click.File


class IntRangeType(_DelegatedDeferredEvaluatorType):
    """:class:`click.IntRange` clone which performs substitution on input."""

    name = "float"
    _delegate_class = click.IntRange


class FloatRangeType(_DelegatedDeferredEvaluatorType):
    """:class:`click.FloatRange` clone which performs substitution on input."""

    name = "float"
    _delegate_class = click.FloatRange


class ChoiceType(_DelegatedDeferredEvaluatorType):
    name = "choice"
    _delegate_class = click.Choice


def multiple_to_layer_ids(layers: int | list[int] | None, document: vp.Document) -> list[int]:
    """Convert multiple-layer CLI argument to list of layer IDs.

    Args:
        layers: value from a :class:`LayerType` argument with accept_multiple=True
        document: target :class:`Document` instance

    Returns:
        List of layer IDs
    """
    if layers is None or layers is LayerType.ALL:
        return sorted(document.ids())
    elif isinstance(layers, list):
        lids = []
        for lid in sorted(layers):
            if document.exists(lid):
                lids.append(lid)
            else:
                logging.info(f"layer {lid} does not exist")
        return lids
    else:
        return []


def single_to_layer_id(
    layer: int | None, document: vp.Document, must_exist: bool = False
) -> int:
    """Convert single-layer CLI argument to layer ID, accounting for the existence of a current
    a current target layer and dealing with default behavior.

    Arg:
        layer: value from a :class:`LayerType` argument
        document: target :class:`Document` instance (for new layer ID)
        must_exists: if True, the function raises a :class:`click.BadParameter` exception

    Returns:
        Target layer ID
    """
    current_target_layer = State.get_current().target_layer_id

    if layer is LayerType.NEW or (layer is None and current_target_layer is None):
        lid = document.free_id()
    elif layer is None:
        lid = State.get_current().target_layer_id
    elif layer < 1:
        raise click.BadParameter(f"layer {layer} is invalid")
    else:
        lid = layer

    if must_exist and lid not in document.layers:
        raise click.BadParameter(f"layer {layer} does not exist")

    return lid


class LayerType(_DeferredEvaluatorType):
    """Interpret values of --layer options.

    If `accept_multiple == True`, comma-separated array of int is accepted or 'all'. Returns
    either a list of IDs or `LayerType.ALL`.

    If `accept_new == True`, 'new' is also accepted, in which case returns `LayerType.NEW`.

    None is passed through, which typically means to use the default behaviour.
    """

    NEW = -1
    ALL = -2

    class _LayerTypeDeferredEvaluator(_DeferredEvaluator):
        def __init__(
            self, text: str, param_name: str, accept_multiple: bool, accept_new: bool
        ):
            super().__init__(text, param_name)
            self._accept_multiple = accept_multiple
            self._accept_new = accept_new

        def evaluate(self, state: State) -> int | list[int]:
            value = state.substitute(self._text)

            if value.lower() == "all":
                if self._accept_multiple:
                    return LayerType.ALL
                else:
                    raise click.BadParameter(
                        f"parameter {self._param_name} must be a single layer and does "
                        "not accept `all`"
                    )
            elif value.lower() == "new":
                if self._accept_new:
                    return LayerType.NEW
                else:
                    raise click.BadParameter(
                        f"parameter {self._param_name} must be an existing layer and "
                        "does not accept `new`"
                    )

            try:
                if self._accept_multiple:
                    id_arr = list(map(int, value.split(",")))
                    for i in id_arr:
                        if i < 1:
                            raise click.BadParameter(f"layer {i} is invalid")
                    return id_arr
                else:
                    return int(value)
            except TypeError:
                raise click.BadParameter(
                    f"unexpected {value!r} of type {type(value).__name__}"
                )
            except ValueError:
                raise click.BadParameter(
                    f"{value!r} is not a valid value for parameter {self._param_name}"
                )

    def __init__(self, accept_multiple: bool = False, accept_new: bool = False):
        super().__init__(accept_multiple=accept_multiple, accept_new=accept_new)

    name = "lid"
    _evaluator_class = _LayerTypeDeferredEvaluator



from __future__ import annotations

import click
import numpy as np

import vpype as vp

from .cli import cli
from .decorators import generator
from .types import IntegerType, LengthType

__all__ = ("random",)


@cli.command(group="Generators")
@click.option(
    "-n", "--count", "n", type=IntegerType(), default=10, help="Number of lines to generate."
)
@click.option(
    "-a",
    "--area",
    nargs=2,
    type=LengthType(),
    default=("10mm", "10mm"),
    help="Dimension of the area in which lines are distributed.",
)
@generator
def random(n: int, area: tuple[float, float]):
    """
    Generate random lines.

    By default, 10 lines are randomly placed in a square with corners at (0, 0) and
    (10mm, 10mm). Use the `--area` option to specify the destination area.
    """

    lines = np.random.rand(n, 2) + 1j * np.random.rand(n, 2)
    lines.real *= area[0]
    lines.imag *= area[1]
    return vp.LineCollection(lines)



from __future__ import annotations

import logging
from typing import Iterable

import numpy as np
from scipy.spatial import KDTree

# REMINDER: anything added here must be added to docs/api.rst
__all__ = ["LineIndex"]


class LineIndex:
    """Wrapper to scipy.spatial.KDTree to facilitate systematic processing of a line
    collection.

    Implementation note: we use the `available` bool array because deleting stuff from the
    index is costly.
    """

    def __init__(self, lines: Iterable[np.ndarray], reverse: bool = False):
        self.lines = [line for line in lines if len(line) > 0]
        self.reverse = reverse
        self._make_index()

    def _make_index(self) -> None:
        logging.info(f"LineIndex: creating index for {len(self.lines)} lines")
        self.available = np.ones(shape=len(self.lines), dtype=bool)

        # create rtree index
        self.index = KDTree(
            np.array([(line[0].real, line[0].imag) for line in self.lines]).reshape(-1, 2)
        )

        # create reverse index
        if self.reverse:
            self.rindex = KDTree(
                np.array([(line[-1].real, line[-1].imag) for line in self.lines]).reshape(
                    -1, 2
                )
            )

    def _reindex(self) -> None:
        self.lines = [line for idx, line in enumerate(self.lines) if self.available[idx]]
        self._make_index()

    def __len__(self) -> int:
        return np.count_nonzero(self.available)

    def __getitem__(self, item):
        return self.lines[item]

    def pop_front(self) -> np.ndarray:
        if len(self) == 0:
            raise RuntimeError
        idx = int(np.argmax(self.available))
        self.available[idx] = False
        return self.lines[idx]

    def pop(self, idx: int) -> np.ndarray | None:
        if not self.available[idx]:
            return None
        self.available[idx] = False
        return self.lines[idx]

    def find_nearest_within(self, p: complex, max_dist: float) -> tuple[int | None, bool]:
        """Find the closest line, assuming a maximum admissible distance.
        Returns a tuple of (idx, reverse), where `idx` may be None if nothing is found.
        `reverse` indicates whether or not a line ending has been matched instead of a start.
        False is always returned if index was created with `reverse=False`.
        """

        ridx = None
        rdist: float | None = 0.0

        while True:
            reindex, idx, dist = self._find_nearest_within_in_index(p, max_dist, self.index)
            if reindex:
                self._reindex()
                continue

            if self.reverse:
                reindex, ridx, rdist = self._find_nearest_within_in_index(
                    p, max_dist, self.rindex
                )
                if reindex:
                    self._reindex()
                    continue
            break

        if self.reverse:
            if idx is None and ridx is None:
                return None, False
            elif idx is not None and ridx is None:
                return idx, False
            elif idx is None and ridx is not None:
                return ridx, True
            elif rdist < dist:  # type: ignore
                return ridx, True
            else:
                return idx, False
        else:
            return idx, False

    def _find_nearest_within_in_index(
        self, p: complex, max_dist: float, index: KDTree
    ) -> tuple[bool, int | None, float | None]:
        """Find nearest in specific index. Return (reindex, idx, dist) tuple, where
        reindex indicates if a reindex is needed.
        """

        # For performance reason, we query only a max of k candidates. In the special case
        # where all distances are not inf and none are available, we might have more than k
        # suitable candidate, so we reindex and loop. Otherwise, we check the query results
        # for availability and not inf and return anything found
        dists, idxs = index.query((p.real, p.imag), k=50, distance_upper_bound=max_dist)
        dists = np.array(dists)

        not_inf = ~np.isinf(dists)
        if np.all(not_inf) and np.all(~self.available[idxs[not_inf]]):
            return True, None, 0

        candidates = self.available[idxs[not_inf]]

        if np.any(candidates):
            idx = np.argmax(candidates)
            return False, idxs[not_inf][idx], dists[not_inf][idx]
        else:
            return False, None, 0

    # noinspection PyUnboundLocalVariable
    def find_nearest(self, p: complex) -> tuple[int, bool]:
        while True:
            idx, dist = self._find_nearest_in_index(p, self.index)
            if self.reverse:
                ridx, rdist = self._find_nearest_in_index(p, self.rindex)

                if ridx is not None and idx is not None:
                    break
            elif idx is not None:
                break
            self._reindex()

        if self.reverse:
            if rdist < dist:
                return ridx, True  # type: ignore
            else:
                return idx, False
        else:
            return idx, False

    def _find_nearest_in_index(self, p: complex, index: KDTree) -> tuple[int | None, float]:
        """Check the N nearest lines, hopefully find one that is active."""

        dists, idxs = index.query((p.real, p.imag), k=100)
        for dist, idx in zip(dists, idxs):
            if ~np.isinf(dist) and self.available[idx]:
                return idx, dist

        return None, 0.0



"""File import/export functions.
"""
from __future__ import annotations

import collections
import copy
import dataclasses
import datetime
import math
import pathlib
import re
from typing import Any, Iterable, Iterator, List, TextIO, Union, cast
from xml.etree import ElementTree

import click
import numpy as np
import svgelements
import svgwrite
import svgwrite.base
from multiprocess import Pool
from shapely.geometry import LineString
from svgwrite.extensions import Inkscape

from .config import PaperConfig, PlotterConfig, config_manager
from .metadata import (
    METADATA_DEFAULT_COLOR_SCHEME,
    METADATA_FIELD_COLOR,
    METADATA_FIELD_NAME,
    METADATA_FIELD_PEN_WIDTH,
    METADATA_FIELD_SOURCE,
    METADATA_SVG_ATTRIBUTES_WHITELIST,
    METADATA_SVG_NAMESPACES,
    Color,
)
from .model import Document, LineCollection
from .utils import UNITS

__all__ = [
    "read_svg",
    "read_multilayer_svg",
    "read_svg_by_attributes",
    "write_svg",
    "write_hpgl",
]


class _ComplexStack:
    """Complex number stack implemented with a numpy array"""

    def __init__(self):
        self._alloc = 100
        self._stack = np.empty(shape=self._alloc, dtype=complex)
        self._len = 0

    def __len__(self) -> int:
        return self._len

    def _realloc(self, min_free: int = 1) -> None:
        self._alloc = max(self._alloc * 2, self._len + min_free)
        # noinspection PyTypeChecker
        self._stack.resize(self._alloc, refcheck=False)

    def append(self, c: complex) -> None:
        if self._len == self._alloc:
            self._realloc()
        self._stack[self._len] = c
        self._len += 1

    def extend(self, a: np.ndarray) -> None:
        len_a = len(a)
        if self._len + len_a > self._alloc:
            self._realloc(len_a)
        self._stack[self._len : self._len + len_a] = a
        self._len += len_a

    def ends_with(self, c: complex) -> bool:
        return self._stack[self._len - 1] == c if self._len > 0 else False

    def get(self) -> np.ndarray:
        self._alloc = self._len
        # noinspection PyTypeChecker
        self._stack.resize(self._alloc, refcheck=False)
        return self._stack


_PathType = Union[
    # for actual paths and shapes transformed into paths
    svgelements.Path,
    # for the special case of Polygon and Polylines
    List[Union[svgelements.PathSegment, svgelements.Polygon, svgelements.Polyline]],
]
_PathListType = List[_PathType]


_NAMESPACED_PROPERTY_RE = re.compile(r"{([-a-zA-Z0-9@:%._+~#=/]+)}([a-zA-Z_][a-zA-Z0-9\-_.]*)")


def _extract_metadata_from_element(
    elem: svgelements.Shape, layer_system_props: bool = True
) -> dict[str, Any]:
    """Extracts the metadata from a svgelements element.

    svgelements offers 3 levels of metadata:
    - object attributes (e.g. `elem.stroke`) which are guaranteed to exist and are typed
      (e.g. `svgelements.Color` for `elem.stroke`).
    - values (e.g. `elem.values`) which are computed from the group hierarchy (e.g. they are
      inherited from enclosing `<g>` elements)
    - SVG attributes (e.g. `elem.values["attributes"]`) which are

    The strategy is as follows:
    - attributes are mapped directly to system metadata (i.e. "vp_color" and "vp_pen_width" --
      "vp.name" is handled by read_multilayer_svg())
    - values are whitelisted and added as `"svg_"` namespace
    - XML-namespaced values are also added as `svg_` namespace (e.g. `"svg_inkscape:label"`
    - SVG attributes are disregarded
    """

    metadata: dict[str, Any] = {}

    # layer-level system properties
    if layer_system_props:
        metadata.update(
            {
                METADATA_FIELD_COLOR: Color(elem.stroke),
                METADATA_FIELD_PEN_WIDTH: elem.stroke_width,
            }
        )

    # white-listed root SVG properties
    metadata.update(
        {
            "svg_" + k: v
            for k, v in elem.values.items()
            if k in METADATA_SVG_ATTRIBUTES_WHITELIST
        }
    )

    # name-spaced XML properties
    for k, v in elem.values.items():
        mo = _NAMESPACED_PROPERTY_RE.match(k)
        if mo:
            ns = mo[1]
            if ns in METADATA_SVG_NAMESPACES:
                metadata[f"svg_{METADATA_SVG_NAMESPACES[ns]}_{mo[2]}"] = v
            else:
                metadata["svg_" + k] = v

    return metadata


def _intersect_dict(a: dict[str, Any], b: dict[str, Any]) -> dict[str, Any]:
    return dict(a.items() & b.items())


def _merge_metadata(
    base_metadata: dict[str, Any] | None, additional_metadata: dict[str, Any]
) -> dict[str, Any]:
    """Merge two metadata dictionaries with handling of uninitialized base dictionary."""

    if base_metadata is None:
        base_metadata = additional_metadata
    else:
        base_metadata = _intersect_dict(base_metadata, additional_metadata)
    return base_metadata


def _element_to_paths(elem: svgelements.SVGElement) -> _PathType | None:
    """Convert a SVG element into a path object that can be processed by
    :func:`_flattened_paths_to_line_collection`

    Args:
        elem: the element to convert

    Returns:
        the path object or None if the element should be ignored
    """
    if isinstance(elem, svgelements.Path):
        if len(elem) != 0:
            return elem
    elif isinstance(elem, (svgelements.Polyline, svgelements.Polygon)):
        # Here we add a "fake" path containing just the Polyline/Polygon,
        # to be treated specifically by _convert_flattened_paths.
        path = [svgelements.Move(elem.points[0]), elem]
        if isinstance(elem, svgelements.Polygon):
            path.append(svgelements.Close(elem.points[-1], elem.points[0]))
        return path
    elif isinstance(elem, svgelements.Shape):
        e = svgelements.Path(elem)
        e.reify()  # In some cases the shape could not have reified, the path must.
        if len(e) != 0:
            return e

    return None


def _extract_paths(
    group: svgelements.Group, recursive
) -> tuple[_PathListType, dict[str, Any] | None]:
    """Extract everything from the provided SVG group."""

    if recursive:
        everything = group.select()
    else:
        everything = group
    paths = []
    metadata: dict[str, Any] | None = None
    for elem in everything:
        if hasattr(elem, "values") and elem.values.get("visibility", "") in (
            "hidden",
            "collapse",
        ):
            continue

        path = _element_to_paths(elem)
        if path is None:
            continue
        else:
            paths.append(path)

        # apply union on metadata
        metadata = _merge_metadata(metadata, _extract_metadata_from_element(elem))

    return paths, metadata


def _extract_paths_by_attributes(
    group: svgelements.Group, attributes: Iterable[str]
) -> list[tuple[_PathListType, dict[str, Any] | None]]:
    """Extract everything from the provided SVG group, grouped by the specified attributes.

    The paths are grouped by unique combinations of the provided attributes.

    Args:
        group: SVG group from which to extract paths
        attributes: attributes by which to group paths

    Returns:
        list of tuple containing the list of paths and the associated metadata
    """

    @dataclasses.dataclass
    class _LayerDesc:
        paths: _PathListType = dataclasses.field(default_factory=list)
        metadata: dict[str, Any] | None = None

    attributes = tuple(attributes)
    results: dict[tuple, _LayerDesc] = collections.defaultdict(lambda: _LayerDesc())

    for elem in group.select():
        if hasattr(elem, "values") and elem.values.get("visibility", "") in (
            "hidden",
            "collapse",
        ):
            continue

        key = tuple(elem.values.get(attr, None) for attr in attributes)

        path = _element_to_paths(elem)
        if path is None:
            continue
        else:
            results[key].paths.append(path)

        # apply union on metadata
        results[key].metadata = _merge_metadata(
            results[key].metadata, _extract_metadata_from_element(elem)
        )

    return [(desc.paths, desc.metadata) for desc in results.values()]


def _flattened_paths_to_line_collection(
    paths: _PathListType,
    quantization: float,
    simplify: bool,
    parallel: bool,
    metadata: dict[str, Any] | None = None,
) -> LineCollection:
    """Convert a path list to a :class:`LineCollection`.

    The resulting :class:`vpype.LineCollection` instance's metadata contains all properties
    (as extracted with :func:`_extract_metadata_from_element`) whose value is identical for
    every single item within the group.

    Note: group-level properties are not explicitly added to the metadata since they are
    propagated to enclosed elements by svgelements.

    Args:
        paths: paths to process
        quantization: maximum length of linear elements to approximate curve paths
        simplify: should Shapely's simplify be run on curved elements after quantization
        parallel: enable multiprocessing
        metadata: if provided, metadata to include in the returned
            :class:`vpype.LineCollection` instance

    Returns:
        new :class:`LineCollection` instance containing the converted geometries and the
        provided metadata
    """

    def _process_path(path):
        if len(path) == 0:
            return []

        result = []
        point_stack = _ComplexStack()
        for seg in path:
            # handle cases of zero radius Arc
            if isinstance(seg, svgelements.Arc) and (seg.rx == 0 or seg.ry == 0):
                seg = svgelements.Line(start=seg.start, end=seg.end)

            if isinstance(seg, svgelements.Move):
                if len(point_stack) > 0:
                    result.append(point_stack.get())
                    point_stack = _ComplexStack()

                point_stack.append(complex(seg.end))
            elif isinstance(seg, (svgelements.Line, svgelements.Close)):
                start = complex(seg.start)
                end = complex(seg.end)
                if not point_stack.ends_with(start):
                    point_stack.append(start)
                if end != start:
                    point_stack.append(end)
            elif isinstance(seg, (svgelements.Polygon, svgelements.Polyline)):
                line = np.array(seg.points, dtype=float)
                line = line.view(dtype=complex).reshape(len(line))
                if point_stack.ends_with(line[0]):
                    point_stack.extend(line[1:])
                else:
                    point_stack.extend(line)
            else:
                # This is a curved element that we approximate with small segments
                step = max(2, int(math.ceil(seg.length() / quantization)))
                line = seg.npoint(np.linspace(0, 1, step))

                if simplify:
                    line = np.array(LineString(line).simplify(tolerance=quantization))

                line = line.view(dtype=complex).reshape(len(line))

                if point_stack.ends_with(line[0]):
                    point_stack.extend(line[1:])
                else:
                    point_stack.extend(line)

        if len(point_stack) > 0:
            result.append(point_stack.get())

        return result

    # benchmarking indicated that parallel processing only makes sense if simplify is used
    if parallel:
        with Pool() as p:
            results = p.map(_process_path, paths)
    else:
        results = map(_process_path, paths)

    lc = LineCollection(metadata=metadata)
    for res in results:
        lc.extend(res)
    return lc


def _get_source(file: str | TextIO) -> pathlib.Path | None:
    try:
        path = pathlib.Path(file)  # type: ignore
        if path.exists():
            return path.absolute()
    except TypeError:
        pass

    return None


def read_svg(
    file: str | TextIO,
    quantization: float,
    crop: bool = True,
    simplify: bool = False,
    parallel: bool = False,
    default_width: float | None = None,
    default_height: float | None = None,
) -> tuple[LineCollection, float, float]:
    """Read a SVG file an return its content as a :class:`LineCollection` instance.

    All curved geometries are chopped in segments no longer than the value of *quantization*.
    Optionally, the geometries are simplified using Shapely, using the value of *quantization*
    as tolerance.

    The page size is set based on the ``width`` and ``height`` attributes of the ``<svg>`` tag.
    If these attributes are missing or expressed in percent, ``svgelements`` attempts to use
    the ``viewBox`` attribute instead, or reverts to a 1000x1000px page size. This behaviour
    can be overridden by providing values for the ``default_width`` and ``default_height``
    arguments.

    Args:
        file: path of the SVG file or stream object
        quantization: maximum size of segment used to approximate curved geometries
        crop: crop the geometries to the SVG boundaries
        simplify: run Shapely's simplify on loaded geometry
        parallel: enable multiprocessing (only recommended for ``simplify=True`` and SVG with
            many curves)
        default_width: default width if not provided by SVG or if a percent width is provided
        default_height: default height if not provided by SVG or if a percent height is
            provided

    Returns:
        tuple containing a :class:`LineCollection` with the imported geometries as well as the
        width and height of the SVG
    """

    # default width is for SVG with % width/height
    svg = svgelements.SVG.parse(file, width=default_width, height=default_height)
    paths, metadata = _extract_paths(svg, True)
    lc = _flattened_paths_to_line_collection(paths, quantization, simplify, parallel, metadata)

    if crop:
        lc.crop(0, 0, svg.width, svg.height)

    source = _get_source(file)
    if source:
        lc.set_property(METADATA_FIELD_SOURCE, source)

    return lc, svg.width, svg.height


def read_multilayer_svg(
    file: str | TextIO,
    quantization: float,
    crop: bool = True,
    simplify: bool = False,
    parallel: bool = False,
    default_width: float | None = None,
    default_height: float | None = None,
) -> Document:
    """Read a multilayer SVG file and return its content as a :class:`Document` instance
    retaining the SVG's layer structure and its dimension.

    Each top-level group is considered a layer. All non-group, top-level elements are imported
    in layer 1.

    Groups are matched to layer ID according their `inkscape:label` attribute, their `id`
    attribute or their appearing order, in that order of priority. Labels are stripped of
    non-numeric characters and the remaining is used as layer ID. Lacking numeric characters,
    the appearing order is used. If the label is 0, its changed to 1.

    All curved geometries are chopped in segments no longer than the value of *quantization*.
    Optionally, the geometries are simplified using Shapely, using the value of *quantization*
    as tolerance.

    The page size is set based on the ``width`` and ``height`` attributes of the ``<svg>`` tag.
    If these attributes are missing or expressed in percent, ``svgelements`` attempts to use
    the ``viewBox`` attribute instead, or reverts to a 1000x1000px page size. This behaviour
    can be overridden by providing values for the ``default_width`` and ``default_height``
    arguments.

    Args:
        file: path of the SVG file or stream object
        quantization: maximum size of segment used to approximate curved geometries
        crop: crop the geometries to the SVG boundaries
        simplify: run Shapely's simplify on loaded geometry
        parallel: enable multiprocessing (only recommended for ``simplify=True`` and SVG with
            many curves)
        default_width: default width if not provided by SVG or if a percent width is provided
        default_height: default height if not provided by SVG or if a percent height is
            provided

    Returns:
         :class:`Document` instance with the imported geometries and its page size set the the
         SVG dimensions
    """

    svg = svgelements.SVG.parse(file, width=default_width, height=default_height)
    document = Document(metadata=_extract_metadata_from_element(svg, False))

    # non-group top level elements are loaded in layer 1
    paths, metadata = _extract_paths(svg, False)
    lc = _flattened_paths_to_line_collection(paths, quantization, simplify, parallel, metadata)
    if not lc.is_empty():
        document.add(lc, 1)
        document.layers[1].metadata = lc.metadata

    def _find_groups(group: svgelements.Group) -> Iterator[svgelements.Group]:
        for elem in group:
            if isinstance(elem, svgelements.Group):
                yield elem

    for i, g in enumerate(_find_groups(svg)):
        # noinspection HttpUrlsUsage
        layer_name = g.values.get("{http://www.inkscape.org/namespaces/inkscape}label", None)

        # compute a decent layer ID
        lid_str = re.sub("[^0-9]", "", layer_name or "")
        if not lid_str:
            lid_str = re.sub("[^0-9]", "", g.values.get("id") or "")
        if lid_str:
            lid = int(lid_str)
            if lid == 0:
                lid = 1
        else:
            lid = i + 1

        paths, metadata = _extract_paths(g, True)
        lc = _flattened_paths_to_line_collection(
            paths, quantization, simplify, parallel, metadata
        )

        if not lc.is_empty():
            # deal with the case of layer 1, which may already be initialized with top-level
            # paths
            if lid in document.layers:
                metadata = _intersect_dict(document.layers[lid].metadata, lc.metadata)
            else:
                metadata = lc.metadata

            document.add(lc, lid)
            document.layers[lid].metadata = metadata
            document.layers[lid].set_property(METADATA_FIELD_NAME, layer_name)

    document.page_size = (svg.width, svg.height)

    if crop:
        document.crop(0, 0, svg.width, svg.height)

    # Because of how svgelements works, all the <svg> level metadata is propagated to every
    # nested tag. As a result, we need to subtract global properties from the layer ones.
    for layer in document.layers.values():
        layer.metadata = dict(layer.metadata.items() - document.metadata.items())

    source = _get_source(file)
    if source:
        document.add_to_sources(source)

    return document


def read_svg_by_attributes(
    file: str | TextIO,
    attributes: Iterable[str],
    quantization: float,
    crop: bool = True,
    simplify: bool = False,
    parallel: bool = False,
    default_width: float | None = None,
    default_height: float | None = None,
) -> Document:
    """Read a SVG file by sorting geometries by unique combination of provided attributes.

    All curved geometries are chopped in segments no longer than the value of *quantization*.
    Optionally, the geometries are simplified using Shapely, using the value of *quantization*
    as tolerance.

    The page size is set based on the ``width`` and ``height`` attributes of the ``<svg>`` tag.
    If these attributes are missing or expressed in percent, ``svgelements`` attempts to use
    the ``viewBox`` attribute instead, or reverts to a 1000x1000px page size. This behaviour
    can be overridden by providing values for the ``default_width`` and ``default_height``
    arguments.

    Args:
        file: path of the SVG file or stream object
        attributes: attributes by which the object should be sorted
        quantization: maximum size of segment used to approximate curved geometries
        crop: crop the geometries to the SVG boundaries
        simplify: run Shapely's simplify on loaded geometry
        parallel: enable multiprocessing (only recommended for ``simplify=True`` and SVG with
            many curves)
        default_width: default width if not provided by SVG or if a percent width is provided
        default_height: default height if not provided by SVG or if a percent height is
            provided

    Returns:
         :class:`Document` instance with the imported geometries and its page size set the the
         SVG dimensions
    """

    svg = svgelements.SVG.parse(file, width=default_width, height=default_height)
    document = Document(metadata=_extract_metadata_from_element(svg, False))

    for paths, metadata in _extract_paths_by_attributes(svg, attributes):
        lc = _flattened_paths_to_line_collection(
            paths, quantization, simplify, parallel, metadata
        )

        if not lc.is_empty():
            lid = document.free_id()
            document.add(lc, lid)
            document.layers[lid].metadata = lc.metadata

    document.page_size = (svg.width, svg.height)

    if crop:
        document.crop(0, 0, svg.width, svg.height)

    # Because of how svgelements works, all the <svg> level metadata is propagated to every
    # nested tag. As a result, we need to subtract global properties from the layer ones.
    for layer in document.layers.values():
        layer.metadata = dict(layer.metadata.items() - document.metadata.items())

    source = _get_source(file)
    if source:
        document.add_to_sources(source)

    return document


_WRITE_SVG_RESTORE_EXCLUDE_LIST = (
    "svg_display",
    "svg_visibility",
    "svg_stroke",
    "svg_stroke-width",
    "svg_inkscape_groupmode",  # handled by svgwrite
    "svg_inkscape_label",  # handled by system field
)


def _restore_metadata(elem: svgwrite.base.BaseElement, metadata: dict[str, Any]) -> None:
    for prop, val in metadata.items():
        if prop.startswith("svg_") and prop not in _WRITE_SVG_RESTORE_EXCLUDE_LIST:
            parts = prop[4:].split("_")
            if len(parts) == 2 and parts[0] in METADATA_SVG_NAMESPACES.values():
                elem.attribs[":".join(parts)] = str(val)
            elif len(parts) == 1:
                elem.attribs[parts[0]] = str(val)


# noinspection HttpUrlsUsage
def write_svg(
    output: TextIO,
    document: Document,
    page_size: tuple[float, float] | None = None,
    center: bool = False,
    source_string: str = "",
    layer_label_format: str | None = None,
    show_pen_up: bool = False,
    color_mode: str = "default",
    use_svg_metadata: bool = False,
    set_date: bool = True,
) -> None:
    """Create a SVG from a :py:class:`Document` instance.

    If no page size is provided (or (0, 0) is passed), the SVG generated has bounds tightly
    fitted around the geometries. Otherwise the provided size (in pixel) is used. The width
    and height is capped to a minimum of 1 pixel.

    By default, no translation is applied on the geometry. If `center=True`, geometries are
    moved to the center of the page.

    No scaling or rotation is applied to geometries.

    Layers are named after the `vp_name` system property if it exists, or with their layer ID
    otherwise. This can be overridden with the  ``layer_label_format`` parameter, which may
    contain a C-style format specifier such as `%d` which will be replaced by the layer number.

    Optionally, metadata properties prefixed with ``svg_`` (typically extracted from an input
    SVG with the :ref:`cmd_read` command) may be used as group attributes with
    ``use_svg_metadata=True``.

    The color of the layer is determined by the ``color_mode`` parameter.

        - ``"default"``: use ``vp_color`` system property and reverts to the default coloring
          scheme
        - ``"none"``: no formatting (black)
        - ``"layer"``: one color per layer based on the default coloring scheme
        - ``"path"``: one color per path based on the default coloring scheme

    The pen width is set to the `vp_pen_width` system property if it exists.

    For previsualisation purposes, pen-up trajectories can be added to the SVG using the
    ``show pen_up`` argument.

    Args:
        output: text-mode IO stream where SVG code will be written
        document: geometries to be written
        page_size: if provided, overrides document.page_size
        center: center geometries on page before export
        source_string: value of the `source` metadata
        layer_label_format: format string for layer label naming
        show_pen_up: add paths for the pen-up trajectories
        color_mode: "default" (system property), "none" (no formatting), "layer" (one color per
            layer), "path" (one color per path)
        use_svg_metadata: apply ``svg_``-prefixed properties as SVG attributes
        set_date: controls whether the current date and time is set in the SVG's metadata
            (disabling it is useful for auto-generated SVG under VCS)
    """

    # compute bounds
    bounds = document.bounds()
    if bounds is None:
        # empty geometry, we provide fake bounds
        bounds = (0, 0, 1, 1)

    if page_size:
        size = page_size
        tight = page_size == (0.0, 0.0)
    elif document.page_size:
        size = document.page_size
        tight = False
    else:
        size = (bounds[2] - bounds[0], bounds[3] - bounds[1])
        tight = True

    if center:
        corrected_doc = copy.deepcopy(document)
        corrected_doc.translate(
            (size[0] - (bounds[2] - bounds[0])) / 2.0 - bounds[0],
            (size[1] - (bounds[3] - bounds[1])) / 2.0 - bounds[1],
        )
    elif tight:
        corrected_doc = copy.deepcopy(document)
        corrected_doc.translate(-bounds[0], -bounds[1])
    else:
        corrected_doc = document

    # output SVG, width/height are capped to 1px
    capped_size = tuple(max(1, s) for s in size)
    size_cm = tuple(f"{round(s / UNITS['cm'], 8)}cm" for s in capped_size)
    dwg = svgwrite.Drawing(size=size_cm, profile="tiny", debug=False)
    inkscape = Inkscape(dwg)
    dwg.attribs.update({"viewBox": f"0 0 {capped_size[0]} {capped_size[1]}"})
    dwg.attribs.update({f"xmlns:{v}": k for k, v in METADATA_SVG_NAMESPACES.items()})

    if use_svg_metadata:
        _restore_metadata(dwg, document.metadata)

    # add metadata
    metadata = ElementTree.Element("rdf:RDF")
    work = ElementTree.SubElement(metadata, "cc:Work")
    fmt = ElementTree.SubElement(work, "dc:format")
    fmt.text = "image/svg+xml"
    source = ElementTree.SubElement(work, "dc:source")
    source.text = source_string
    if set_date:
        date = ElementTree.SubElement(work, "dc:date")
        date.text = datetime.datetime.now().isoformat()
    dwg.set_metadata(metadata)

    color_idx = 0
    if show_pen_up:
        group = inkscape.layer(label="% pen up trajectories")
        group.attribs["fill"] = "none"
        group.attribs["stroke"] = "black"
        group.attribs["style"] = "display:inline; stroke-opacity: 50%; stroke-width: 0.5"
        group.attribs["id"] = "pen_up_trajectories"

        for layer in corrected_doc.layers.values():
            for line in layer.pen_up_trajectories():
                group.add(
                    dwg.line((line[0].real, line[0].imag), (line[-1].real, line[-1].imag))
                )

        dwg.add(group)

    for layer_id in sorted(corrected_doc.layers.keys()):
        layer = corrected_doc.layers[layer_id]

        if layer_label_format is not None:
            label = layer_label_format % layer_id
        elif layer.property_exists(METADATA_FIELD_NAME):
            label = str(layer.property(METADATA_FIELD_NAME))
        else:
            label = str(layer_id)

        group = inkscape.layer(label=label)
        group.attribs["fill"] = "none"

        color = Color("black")
        if color_mode == "layer" or (
            color_mode == "default" and not layer.property_exists(METADATA_FIELD_COLOR)
        ):
            color = METADATA_DEFAULT_COLOR_SCHEME[
                color_idx % len(METADATA_DEFAULT_COLOR_SCHEME)
            ]
            color_idx += 1
        elif color_mode == "default":
            color = Color(layer.property(METADATA_FIELD_COLOR))

            # we want to avoid a subsequent layer whose color is undefined to have its color
            # affected by whether or not previous layer have their color defined
            color_idx += 1

        group.attribs["stroke"] = color.as_rgb_hex()
        if color.alpha < 255:
            group.attribs["stroke-opacity"] = f"{color.alpha/255:.3f}"
        group.attribs["style"] = "display:inline"
        group.attribs["id"] = f"layer{layer_id}"
        if layer.property_exists(METADATA_FIELD_PEN_WIDTH):
            group.attribs["stroke-width"] = float(
                cast(float, layer.property(METADATA_FIELD_PEN_WIDTH))
            )

        # dump all svg properties as attribute
        if use_svg_metadata:
            _restore_metadata(group, layer.metadata)

        for line in layer:
            if len(line) <= 1:
                continue

            if len(line) == 2:
                path = dwg.line((line[0].real, line[0].imag), (line[1].real, line[1].imag))
            elif line[0] == line[-1]:
                path = dwg.polygon((c.real, c.imag) for c in line[:-1])
            else:
                path = dwg.polyline((c.real, c.imag) for c in line)

            if color_mode == "path":
                path.attribs["stroke"] = METADATA_DEFAULT_COLOR_SCHEME[
                    color_idx % len(METADATA_DEFAULT_COLOR_SCHEME)
                ]
                color_idx += 1
            group.add(path)

        dwg.add(group)

    dwg.write(output, pretty=True)


def _get_hpgl_config(device: str | None, page_size: str) -> tuple[PlotterConfig, PaperConfig]:
    plotter_config = config_manager.get_plotter_config(device)
    if plotter_config is None:
        raise ValueError(f"no configuration available for plotter '{device}'")
    paper_config = plotter_config.paper_config(page_size)
    if paper_config is None:
        raise ValueError(
            f"no configuration available for paper size '{page_size}' with plotter "
            f"'{device}'"
        )

    return plotter_config, paper_config


def write_hpgl(
    output: TextIO,
    document: Document,
    page_size: str,
    landscape: bool,
    center: bool,
    device: str | None,
    velocity: int | None,
    absolute: bool = False,
    quiet: bool = False,
) -> None:
    """Create a HPGL file from the :class:`Document` instance.

    The ``device``/``page_size`` combination must be defined in the built-in or user-provided
    config files or an exception will be raised.

    By default, no translation is applied on the geometry. If `center=True`, geometries are
    moved to the center of the page.

    No scaling or rotation is applied to geometries.

    Args:
        output: text-mode IO stream where SVG code will be written
        document: geometries to be written
        page_size: page size string (it must be configured for the selected device)
        landscape: if True, the geometries are generated in landscape orientation
        center: center geometries on page before export
        device: name of the device to use (the corresponding config must exists). If not
            provided, a default device must be configured, which will be used.
        velocity: if provided, a VS command will be generated with the corresponding value
        absolute: if True, only use absolute coordinates
        quiet: if True, do not print the plotter/paper info strings
    """

    # empty HPGL is acceptable there are no geometries to plot
    if document.is_empty():
        return

    plotter_config, paper_config = _get_hpgl_config(device, page_size)
    if not quiet:
        if plotter_config.info:
            # use of echo instead of print needed for testability
            # https://github.com/pallets/click/issues/1678
            click.echo(plotter_config.info, err=True)
        if paper_config.info:
            click.echo(paper_config.info, err=True)

    # Handle flex paper size.
    # If paper_size is not provided by the config, the paper size is then assumed to be the
    # same a the current page size. In this case, the config should provide paper_orientation
    # since it may not be the same as the document's page size
    paper_size = paper_config.paper_size
    if paper_size is None:
        if document.page_size is not None:
            paper_size = document.page_size
        else:
            raise ValueError(
                "paper size must be set using `read`, `pagesize`, or `layout` command"
            )

    # Ensure paper orientation is correct.
    if paper_config.paper_orientation is not None and (
        (paper_config.paper_orientation == "portrait" and paper_size[0] > paper_size[1])
        or (paper_config.paper_orientation == "landscape" and paper_size[0] < paper_size[1])
    ):
        paper_size = paper_size[::-1]

    # Handle origin location.
    origin_x, origin_y = paper_config.origin_location
    if paper_config.origin_location_reference not in ["topleft", "botleft"]:
        raise ValueError(
            "incorrect value for origin_location_reference: "
            f"{paper_config.origin_location_reference}"
        )
    if paper_config.origin_location_reference == "botleft":
        origin_y = paper_size[1] - origin_y

    # are plotter coordinate placed in landscape or portrait orientation?
    coords_landscape = paper_size[0] > paper_size[1]

    # document preprocessing:
    # - make a copy
    # - deal with orientation mismatch
    # - optionally center on paper
    # - convert to plotter units
    # - crop to plotter limits
    document = copy.deepcopy(document)

    if landscape != coords_landscape:
        document.rotate(-math.pi / 2)
        document.translate(0, paper_size[1])

    if paper_config.rotate_180:
        document.scale(-1, -1)
        document.translate(*paper_size)

    if center:
        bounds = document.bounds()
        if bounds is not None:
            document.translate(
                (paper_size[0] - (bounds[2] - bounds[0])) / 2.0 - bounds[0],
                (paper_size[1] - (bounds[3] - bounds[1])) / 2.0 - bounds[1],
            )

    document.translate(-origin_x, -origin_y)
    unit_per_pixel = 1 / plotter_config.plotter_unit_length
    document.scale(
        unit_per_pixel, -unit_per_pixel if paper_config.y_axis_up else unit_per_pixel
    )
    if paper_config.x_range is not None and paper_config.y_range is not None:
        document.crop(
            paper_config.x_range[0],
            paper_config.y_range[0],
            paper_config.x_range[1],
            paper_config.y_range[1],
        )

    # output HPGL
    def complex_to_str(p: complex) -> str:
        return f"{int(round(p.real))},{int(round(p.imag))}"

    output.write("IN;DF;")
    if velocity is not None:
        output.write(f"VS{int(velocity)};")
    if paper_config.set_ps is not None:
        output.write(f"PS{int(paper_config.set_ps)};")

    first_layer = True
    last_point: complex | None = None
    for layer_id in sorted(document.layers.keys()):
        pen_id = 1 + (layer_id - 1) % plotter_config.pen_count

        # As per #310, we emit a single PU; between layers
        if not first_layer:
            output.write(f"PU;")
        else:
            first_layer = False
        output.write(f"SP{pen_id};")

        for line in document.layers[layer_id]:
            if len(line) < 2:
                continue

            if absolute:
                output.write(
                    f"PU{complex_to_str(line[0])};PD"
                    + ",".join(complex_to_str(p) for p in line[1:])
                    + ";"
                )
            else:
                # snap all points of the line to the HPGL grid, such as to avoid any aliasing
                # when computing the diff between points
                line = np.round(line)

                if last_point is None:
                    output.write(f"PU{complex_to_str(line[0])};PR;")
                else:
                    output.write(f"PU{complex_to_str(line[0] - last_point)};")

                output.write("PD" + ",".join(complex_to_str(p) for p in np.diff(line)) + ";")
                last_point = line[-1]

    if not absolute:
        output.write("PA;")
    output.write(f"PU{paper_config.final_pu_params if paper_config.final_pu_params else ''};")
    output.write("SP0;IN;\n")
    output.flush()
